# -*- coding: utf-8 -*-
'''
  fastmat/Product.py
 -------------------------------------------------- part of the fastmat package

  Product of fastmat matrices.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.cmath cimport _conjugate
from .core.types cimport *


################################################################################
################################################## class Product
cdef class Product(Matrix):

    ############################################## class methods
    def __init__(self, *matrices, **options):
        '''Initialize Matrix instance'''

        # evaluate options passed to Product
        debug = options.get('debug', False)

        # initialize product content to [1]
        lstFactors = []
        self._scalar = np.int8(1)

        # determine data type of matrix: store datatype in immutable object to
        # allow access from within subfunctions (python2/3 compatible)
        datatype = [self._scalar.dtype]

        def __promoteType(dtype):
            datatype[0] = np.promote_types(datatype[0], dtype)

        # add all product terms, direct and nested
        def __addFactors(factors):
            for factor in factors:
                if np.isscalar(factor):
                    # explicit type promotion: avoid type shortening of scalars
                    __promoteType(np.array(factor).dtype)
                    self._scalar = np.inner(self._scalar, factor)
                    continue

                if not isinstance(factor, Matrix):
                    raise TypeError("Product: Term not scalar nor Matrix.")

                if isinstance(factor, Product):
                    if factor._scalar != 1:
                        self._scalar = self._scalar * factor._scalar

                    __addFactors(factor.content)
                else:
                    # store fastmat-matrix-content: determine data type
                    # -> promotion of factor types
                    __promoteType(factor.dtype)
                    lstFactors.append(factor)
        __addFactors(matrices)
        dtype = datatype[0]

        # handle type expansion with default depending on matrix type
        # default: expand small types due to accumulation during transforms
        # skip by specifying `typeExpansion=None` or override with `~=...`
        typeExpansion = options.get('typeExpansion', safeTypeExpansion(dtype))
        dtype = (dtype if typeExpansion is None
                 else np.promote_types(dtype, typeExpansion))

        # sanity check of the supplied amount of product terms
        if len(lstFactors) < 1:
            raise ValueError("Product has no terms.")

        # iterate elements and check if their numN fit the previous numM
        cdef intsize numN = lstFactors[0].numN
        cdef intsize numM = lstFactors[0].numM
        cdef int ii
        for ii in range(1, len(lstFactors)):
            factor = lstFactors[ii]
            if factor.numN != numM:
                raise ValueError(
                    "Product: Dimension mismatch for term %d [%dx%d]" %(
                        ii, numN, numM))
            numM = factor.numM

        # force scalar datatype to match matrix datatype (calculation accuracy)
        self._scalar = self._scalar.astype(dtype)

        # also, make factor list immutable
        self._content = tuple(lstFactors)

        # set properties of matrix
        self._initProperties(numN, numM, dtype)

        if debug:
            print("fastmat.Product instance %12x containing:" %(id(self)))
            if self._scalar != 1:
                print("  [0]: scalar %s" %(self._scalar))
            for ii, factor in enumerate(lstFactors):
                print("  [%d]: %s" %(ii, factor.__repr__()))

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        # regular product except the last term is just term._getCol
        cdef int cnt = len(self._content)
        cdef int ii                     # index (0 .. cnt - 1)
        cdef int iii = cnt - 1          # index (cnt - 1 .. 0)
        cdef np.ndarray arrRes = self._content[iii].getCol(idx)

        # use inner for element-wise scalar mul as inner does type promotion
        if self._scalar != 1:
            arrRes = np.inner(arrRes, self._scalar)
        else:
            arrRes = arrRes.astype(np.promote_types(arrRes.dtype, self.dtype))

        for ii in range(1, cnt):
            iii = cnt - 1 - ii
            arrRes = self._content[iii].forward(arrRes)

        return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        # regular product w/ backward except the last term is just term._getRow
        cdef int cnt = len(self._content)
        cdef int ii = 0
        cdef np.ndarray arrRes = _conjugate(self._content[ii].getRow(idx))

        # use inner for element-wise scalar mul as inner does type promotion
        if self._scalar != 1:
            if np.iscomplex(self._scalar):
                arrRes = np.inner(arrRes, self._scalar.conjugate())
            else:
                arrRes = np.inner(arrRes, self._scalar)
        else:
            arrRes = arrRes.astype(np.promote_types(arrRes.dtype, self.dtype))

        for ii in range(1, cnt):
            arrRes = self._content[ii].backward(arrRes)

        # don't forget to return the conjugate as we use the backward
        return _conjugate(arrRes)

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = len(self._content)
        return (complexity, complexity)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''

        cdef int cnt = len(self._content)
        cdef int ii                     # index (0 .. cnt - 1)
        cdef int iii = cnt - 1          # index (cnt - 1 .. 0)
        cdef np.ndarray arrRes = arrX

        # use inner for element-wise scalar mul as inner does type promotion
        if self._scalar != 1:
            arrRes = np.inner(arrRes, self._scalar)
        else:
            arrRes = arrRes.astype(np.promote_types(arrRes.dtype, self.dtype))

        for ii in range(0, cnt):
            iii = cnt - 1 - ii
            arrRes = self._content[iii].forward(arrRes)

        return arrRes

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''

        cdef int cnt = len(self._content)
        cdef int ii
        cdef np.ndarray arrRes = arrX

        # use inner for element-wise scalar mul as inner does type promotion
        if self._scalar != 1:
            if np.iscomplex(self._scalar):
                arrRes = np.inner(arrRes, self._scalar.conjugate())
            else:
                arrRes = np.inner(arrRes, self._scalar)
        else:
            arrRes = arrRes.astype(np.promote_types(arrRes.dtype, self.dtype))

        for ii in range(0, cnt):
            arrRes = self._content[ii].backward(arrRes)

        return arrRes

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef ii, cnt = len(self._content)
        cdef Matrix term
        cdef np.ndarray arrRes

        dtype = np.promote_types(np.float64, self.dtype)

        arrRes = np.inner(
            self._content[cnt - 1].reference(), self._scalar.astype(dtype))

        for ii in range(1, cnt):
            term = self._content[cnt - ii - 1]
            arrRes = term.reference().dot(arrRes)

        return arrRes

    def _forwardReferenceInit(self):
        self._forwardReferenceMatrix = []
        for ii, term in enumerate(self._content):
            self._forwardReferenceMatrix.append(term.reference())

    def _forwardReference(self, arrX):
        '''Calculate the forward transform by non-fastmat means.'''

        # check if operation list initialized. If not, then do it!
        if not isinstance(self._forwardReferenceMatrix, list):
            self._forwardReferenceInit()

        # perform operations list
        arrRes = arrX
        for ii in range(len(self._content), 0, -1):
            arrRes = self._forwardReferenceMatrix[ii - 1].dot(arrRes)

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 7,
                TEST.NUM_M      : TEST.Permutation([10, TEST.NUM_N]),
                'mType1'        : TEST.Permutation(TEST.ALLTYPES),
                'mType2'        : TEST.Permutation(TEST.FEWTYPES),
                'sType'         : TEST.Permutation(TEST.ALLTYPES),
                'arr1'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M)
                }),
                'arr2'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType2',
                    TEST.SHAPE  : (TEST.NUM_M , TEST.NUM_N)
                }),
                'arr3'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : (TEST.NUM_N , TEST.NUM_M)
                }),
                'num4'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'sType',
                    TEST.SHAPE  : (1,)
                }),
                TEST.OBJECT     : Product,
                TEST.INITARGS   : (lambda param: [param['num4']()[0],
                                                  Matrix(param['arr1']()),
                                                  Matrix(param['arr2']()),
                                                  Matrix(param['arr3']())]),
                'strType'       : (lambda param: TEST.TYPENAME[param['sType']]),
                TEST.NAMINGARGS: dynFormat("%s*%s*%s*%s",
                                           'strType', 'arr1', 'arr2', 'arr3'),
                TEST.TOL_POWER  : 3.
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        from .Fourier import Fourier
        from .Hadamard import Hadamard
        from .Eye import Eye
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c:
                                   Product(Hadamard(c), Fourier(2 ** c))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_STEP : (lambda c: c + 1)
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Product(*([Eye(2 ** c)] * 2 ** c)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Product (\texttt{fastmat.Product})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
\[\bm M = \prod\limits_i \bm A_i \]
where the $A_{i}$ can be fast transforms of \emph{any} type.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the product terms',
                            'A = fm.Circulant(x_A)',
                            'B = fm.Circulant(x_B)',
                            '',
                            '# construct the product',
                            'M = fm.Product(A.H, B)',
                            caption=r"""
Assume we have two circulant matrices $\bm A$ and $\bm B$. Then we define
\[\bm M = \bm A_c^\herm \bm B_c.\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
The \emph{forward} and \emph{solve} benchmarks were performed on a matrix
$\bm P = \bm \Hs_k \cdot \bm \Fs_{2^k}$; so $n = 2^k$
whereas the \emph{overhead} benchmark was performed on a
Produkt of $2^k$ Identity matrices $\bm I_{2^k}$; so $n = 2^k$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD()
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/Blocks.py
 -------------------------------------------------- part of the fastmat package

  Block matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------

  TODO:
    - Blocks should simply skip all Zero Matrices (flag them as "None")?
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix


################################################################################
################################################## class Blocks
cdef class Blocks(Matrix):

    ############################################## class methods
    def __init__(self, arrMatrices):
        '''Initialize Matrix instance with a list of child matrices'''
        if not isinstance(arrMatrices, list):
            raise ValueError("Blocks: Not a nested list of fastmat matrices.")

        if len(arrMatrices) < 1:
            raise ValueError("Blocks: Contains no matrices.")

        cdef intsize numN = 0, numM = 0
        cdef intsize ii, rr, cc
        cdef tuple row, firstRow
        cdef Matrix term

        # initialize sizes and number of rows / cols
        self._numRows = len(arrMatrices)
        self._numCols = len(arrMatrices[0])
        dataType = np.int8

        # generate transposed block structure, force tuples for backward()
        cdef list lst = [tuple(arrMatrices[rr]) for rr in range(self._numRows)]
        self._rows = tuple(lst)
        lst = [[] for _ in range(self._numCols)]
        firstRow = self._rows[0]

        # extract list of row heights and column widths
        self._rowN = tuple([row[0].numN for row in self._rows])
        self._colM = tuple([term.numM for term in firstRow])

        # enumerate rows
        for rr in range(self._numRows):
            row = self._rows[rr]

            # get number of rows from first elements in columns
            numN += row[0].numN

            # check for presence of enough blocks
            if len(row) != self._numCols:
                raise ValueError(
                    "Blocks.row(%d) has incompatible number of entries" %(rr))

            # enumerate columns
            for cc in range(self._numCols):
                term = row[cc]

                # check for matching column height and width
                if term.numN != row[0].numN:
                    raise ValueError(
                        ("Blocks[%d,%d] with shape %s is incompatible with " +
                         "row height (%d,:)") %(
                            rr, cc, str(term.shape), row[0].numN))

                if term.numM != firstRow[cc].numM:
                    raise ValueError(
                        ("Blocks[%d,%d] with shape %s is incompatible with " +
                         "column width (:,%d)") %(
                            rr, cc, str(term.shape), firstRow[cc].numM))

                # first run: get stats and update dimension of
                # Blocks from first column's entries
                if rr == 0:
                    numM += term.numM

                # build transposed copy of blocks to work on
                # in backward
                lst[cc].append(term)

                # determine necessary output data type by
                # applying type promotion
                dataType = np.promote_types(dataType, term.dtype)

        # convert generated transposed array to tuple of tuples
        for cc in range(self._numCols):
            lst[cc] = tuple(lst[cc])

        self._cols = tuple(lst)

        # build a flat list of all nested matrices in _content
        self._content = tuple([item for row in self._rows for item in row])

        # set properties of matrix
        self._initProperties(
            numN, numM, dataType,
            cythonCall=True,
            widenInputDatatype=True
        )

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexityFwd, complexityBwd
        cdef Matrix item

        complexityFwd = complexityBwd = sum(self._colM) + sum(self._rowN)
        for item in self:
            complexityFwd += item.numN + item.numM
            complexityBwd += item.numM + item.numN

        return (complexityFwd, complexityBwd)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the forward transform of this matrix'''
        cdef np.ndarray viewOut, arrOut
        cdef Matrix term
        cdef tuple row, viewRows, viewCols
        cdef intsize idxN, idxM, rr, cc
        cdef list lst

        # generate views into output array
        lst = [None] * self._numRows
        idxN = 0
        for rr in range(self._numRows):
            lst[rr] = arrRes[idxN:(idxN + self._rowN[rr])]
            idxN += self._rowN[rr]

        viewRows = tuple(lst)

        # generate views into input array
        lst = [None] * self._numCols
        idxM = 0
        for cc in range(self._numCols):
            lst[cc] = arrX[idxM:(idxM + self._colM[cc])]
            idxM += self._colM[cc]

        viewCols = tuple(lst)

        # do the trick
        for rr in range(self._numRows):
            row = self._rows[rr]
            viewOut = viewRows[rr]

            viewOut[:] = row[0].forward(viewCols[0])
            for cc in range(1, self._numCols):
                viewOut += row[cc].forward(viewCols[cc])

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the backward transform of this matrix'''
        cdef np.ndarray viewOut
        cdef Matrix term
        cdef tuple col, viewRows, viewCols
        cdef intsize idxN, idxM, rr, cc
        cdef list lst

        # generate views into output array
        lst = [None] * self._numCols
        idxM = 0
        for cc in range(self._numCols):
            lst[cc] = arrRes[idxM:(idxM + self._colM[cc])]
            idxM += self._colM[cc]

        viewCols = tuple(lst)

        # generate views into input array
        lst = [None] * self._numRows
        idxN = 0
        for rr in range(self._numRows):
            lst[rr] = arrX[idxN:(idxN + self._rowN[rr])]
            idxN += self._rowN[rr]

        viewRows = tuple(lst)

        # do the trick
        for cc in range(self._numCols):
            col = self._cols[cc]
            viewOut = viewCols[cc]
            viewOut[:] = col[0].backward(viewRows[0])

            for rr in range(1, self._numRows):
                viewOut += col[rr].backward(viewRows[rr])

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes
        cdef Matrix term
        cdef tuple row
        cdef intsize idxN = 0, idxM = 0

        arrRes = np.empty((self.numN, self.numM), dtype=self.dtype)

        cdef intsize rr, tt
        for rr in range(self._numRows):
            row = self._rows[rr]
            idxM = 0
            for tt in range(self._numCols):
                term = row[tt]
                arrRes[idxN:(idxN + term.numN), idxM:(idxM + term.numM)] = \
                    term._reference()
                idxM += term.numM

            idxN += term.numN

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                'size'          : 4,
                TEST.NUM_N      : (lambda param: param['size'] * 2),
                TEST.NUM_M      : TEST.NUM_N,
                'mType1'        : TEST.Permutation(TEST.ALLTYPES),
                'mType2'        : TEST.Permutation(TEST.ALLTYPES),
                'arr1'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : ('size', 'size')
                }),
                'arr2'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType2',
                    TEST.SHAPE  : ('size', 'size')
                }),
                'arr3'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType2',
                    TEST.SHAPE  : ('size', 'size')
                }),
                'arr4'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : ('size', 'size')
                }),
                TEST.INITARGS   : (lambda param : [[[Matrix(param['arr1']()),
                                                     Matrix(param['arr2']())],
                                                    [Matrix(param['arr3']()),
                                                     Matrix(param['arr4']())]]
                                                   ]),
                TEST.OBJECT     : Blocks,
                'strType1'      : (lambda param:
                                   TEST.TYPENAME[param['mType1']]),
                'strType2'      : (lambda param:
                                   TEST.TYPENAME[param['mType2']]),
                TEST.NAMINGARGS : dynFormat("[%s,%s],[%s,%s]:(%dx%d) each",
                                            'strType1', 'strType2',
                                            'strType2', 'strType1',
                                            'size', 'size')
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        from .Circulant import Circulant
        from .Diag import Diag
        from .Eye import Eye
        from .Fourier import Fourier
        return {
            BENCH.FORWARD: {
                BENCH.FUNC_GEN  : (lambda c: Blocks(
                    [[Circulant(np.random.randn(c)),
                      Circulant(np.random.randn(c))],
                     [Fourier(c), Diag(np.random.randn(c))]])),
                BENCH.FUNC_SIZE : (lambda c: 2 * c)
            },
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Blocks([[Eye(2 ** c)] * 4] * 4)),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c * 4)
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Block Matrix (\texttt{fastmat.Blocks})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
\[\bm M = \left( \bm A_{i,j}\right)_{i,j},\] where the $\bm A_{i,j}$ can be
fast transforms of \emph{any} type.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the blocks',
                            'A = fm.Circulant(x_A)',
                            'B = fm.Circulant(x_B)',
                            'C = fm.Fourier(n)',
                            'D = fm.Diag(x_D)'
                            '',
                            '# define the block',
                            '# matrix row-wise',
                            'M = fm.Blocks([[A,B],[C,D]])',
                            caption=r"""
Assume we have two circulant matrices $\bm A$ and $\bm B$, an $N$-dimensional
Fourier matrix $\bm C$ and a diagonal matrix $\bm D$. Then we define
\[
    \bm M =
        \left(\begin{array}{cc}
                \bm A & \bm B \\ \bm C & \bm D
              \end{array}\right).\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks',
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD(doc=r"""
$\bm B = \begin{pmatrix}
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} \\
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} \\
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} \\
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k}
\end{pmatrix}$, $n = 2^{k+2}$ for $k \in \N$""")
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/Partial.py
 -------------------------------------------------- part of the fastmat package

  Partial matrices.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.types cimport *
from .core.cmath cimport _arrZero

################################################################################
################################################## class Partial
cdef class Partial(Matrix):

    ############################################## class properties
    # indicesN, indicesM - Property (read-only)
    # Define the support of the base matrix which defines the partial.
    property indicesN:
        def __get__(self):
            return self._indicesN

    property indicesM:
        def __get__(self):
            return self._indicesM

    ############################################## class methods
    def __init__(
        self,
        mat,
        N=None,
        M=None
    ):
        '''Initialize Matrix instance'''

        # initialize matrix for full support (used anyway for checking)
        if not isinstance(mat, Matrix):
            raise ValueError("Partial: fastmat Matrix required.")
        self._indicesM = np.arange(mat.numM)
        self._indicesN = np.arange(mat.numN)
        self._content = (mat, )

        # check if anything needs to be done in N- or M-dimension
        # store support indices in N- and M- dimension if needed
        self._pruneN = False
        if N is not None:
            N = np.array(N)
            if (len(N) != mat.numN) or (np.sum(N - self._indicesN) != 0):
                if np.any((N >= mat.numN) | (N < 0)):
                    raise ValueError(
                        "Partial: A row index exceed matrix dimensions.")

                self._indicesN = N
                self._pruneN = True

        self._pruneM = False
        if M is not None:
            M = np.array(M)
            if (len(M) != mat.numM) or (np.sum(M - self._indicesM) != 0):
                if np.any((M >= mat.numM) | (M < 0)):
                    raise ValueError(
                        "Partial: A column index exceeds matrix dimensions.")

            self._indicesM = M
            self._pruneM = True

        # set properties of matrix
        self._initProperties(
            len(self._indicesN), len(self._indicesM), mat.dtype)

    def __repr__(self):
        '''
        Return a string representing this very class instance.
        The __repr__() of the nested matrix is extended by an info about
        the applied reduction if the Partial itself is not subclassed
        further.
        '''
        # determine if subclassed
        if type(self) == Partial:
            return "<%s[%dx%d](%s[%dx%d]):0x%12x>" %(
                self.__class__.__name__,
                self.numN, self.numM,
                self._content[0].__class__.__name__,
                self._content[0].numN, self._content[0].numM,
                id(self))
        else:
            return super(Partial, self).__repr__()

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        cdef intsize idxM = self._indicesM[idx] if self._pruneM else idx
        return (self._content[0].getCol(idxM)[self._indicesN] if self._pruneN
                else self._content[0].getCol(idxM))

    cpdef np.ndarray _getRow(self, intsize idx):
        cdef intsize idxN = self._indicesN[idx] if self._pruneN else idx
        return (self._content[0].getRow(idxN)[self._indicesM] if self._pruneM
                else self._content[0].getRow(idxN))

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef Matrix M = self._content[0]
        cdef float complexityFwd = 0.
        cdef float complexityBwd = 0.
        if self._pruneM:
            complexityFwd += M.numM + self.numM
            complexityBwd += self.numN

        if self._pruneN:
            complexityBwd += M.numN + self.numN
            complexityFwd += self.numM

        return (complexityFwd, complexityBwd)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''

        cdef np.ndarray arrInput

        if self._pruneM:
            arrInput = _arrZero(
                2, self._content[0].numM, arrX.shape[1], _getNpType(arrX))
            arrInput[self._indicesM, :] = arrX
        else:
            arrInput = arrX

        return (self._content[0].forward(arrInput)[self._indicesN, :]
                if self._pruneN else self._content[0].forward(arrInput))

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''

        cdef np.ndarray arrInput

        if self._pruneN:
            arrInput = _arrZero(
                2, self._content[0].numN, arrX.shape[1], _getNpType(arrX))
            arrInput[self._indicesN, :] = arrX
        else:
            arrInput = arrX

        return (self._content[0].backward(arrInput)[self._indicesM, :]
                if self._pruneM else self._content[0].backward(arrInput))

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrFull = self._content[0].reference()
        return arrFull[self._indicesN, :][:, self._indicesM]

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        from .Hadamard import Hadamard
        return {
            TEST.COMMON: {
                'order'         : 4,
                'num_N'         : (lambda param: 2 ** param['order']),
                'num_M'         : 'num_N',
                TEST.NUM_N      : (lambda param: len(param['subRows'])),
                TEST.NUM_M      : (lambda param: len(param['subCols'])),

                'subCols'       : TEST.Permutation([np.array([1, 2, 3, 11, 12]),
                                                    np.array([6])]),
                'subRows'       : TEST.Permutation([np.array([7, 8, 9, 13]),
                                                    np.array([10])]),
                TEST.INITARGS   : (lambda param: [Hadamard(param['order']),
                                                  param['subRows'],
                                                  param['subCols']]),
                TEST.OBJECT     : Partial,
                TEST.NAMINGARGS : dynFormat("Hadamard(%d)->%dx%d",
                                            'order', TEST.NUM_N, TEST.NUM_M)
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        from .Eye import Eye
        return {
            BENCH.FORWARD: {
                BENCH.FUNC_GEN  : (lambda c: Partial(
                    Eye(2 * c), N=np.arange(c), M=np.arange(c)))
            },
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c:
                                   Partial(Eye(2 ** c), np.arange(2 ** c)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Partial Transform (\texttt{fastmat.Partial})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
Let $I \subset \{1,\dots,n\}$ and $J \subset \{1,\dots,m\}$ index sets and
$\bm M \in \C^{n \times m}$ a linear transform. Then the partial transform
$\bm M_{I,J}$ is defined as
\[\bm x \in \C^m \mapsto (\bm M_J \cdot \bm x_J)_{i \in I}.\]
In other words, we select the rows $I$ of $\bm M$ and columns J of $\bm M$ and
rows $J$ of $\bm x$.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# define the index set',
                            'a = np.arange(n)',
                            'am = np.mod(a, 2)',
                            "b = np.array(am, dtype='bool')",
                            'I = a[b]',
                            '',
                            '# construct the partial transform',
                            'M = fm.Partial(F, I)',
                            caption=r"""
Let $\bm{\mathcal{F}}$ be the $n$-dimensional Fourier matrix. And let $I$ be the
set of odd integers. Then we define a partial transform as
\[\bm M = \bm{\mathcal{F}}_I\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
The forward projection benchmarks were performed on a partial Hadamard matrix
$\bm P = \bm \Hs_{2^n,\{1,\dots,n\}}$ while the runtime performance benchmark
was performed on a partial Identity matrix.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD()
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/Kron.pyx
 -------------------------------------------------- part of the fastmat package

  Kronecker Product of matrices.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix, profileUpdate
from .Product cimport Product
from .core.types cimport *
from .core.cmath cimport _arrReshape, _arrEmpty

################################################################################
################################################## class Kron
cdef class Kron(Matrix):

    ############################################## class methods
    def __init__(self, *matrices, **options):
        '''Initialize Matrix instance with a list of child matrices'''

        cdef int ff, factorCount = len(matrices)
        cdef intsize numN = 1
        cdef Matrix factor

        #check the number of matrices
        self._content = tuple(matrices)
        if factorCount < 2:
            raise ValueError("Kronecker: Product must have at least two terms")

        # determine total size and data type of matrix
        dtype = np.int8

        for ff in range(factorCount):
            factor = self._content[ff]

            #check for symmetry of matrices
            if (factor.numN != factor.numM):
                raise ValueError("Kronecker: Product terms must be symmetric")

            # acknowledge size and data type of factor
            numN *= self._content[ff].numN
            dtype = np.promote_types(dtype, factor.dtype)

        # determine dimensions of all factor terms in one tuple
        self._dims = tuple([factor.numN for factor in self._content])

        # handle type expansion with default depending on matrix type
        # default: expand small types due to accumulation during transforms
        # skip by specifying `typeExpansion=None` or override with `~=...`
        typeExpansion = options.get('typeExpansion', safeTypeExpansion(dtype))
        dtype = (dtype if typeExpansion is None
                 else np.promote_types(dtype, typeExpansion))

        # set properties of matrix
        self._initProperties(numN, numN, dtype, widenInputDatatype=True)

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        # perform index decomposition
        cdef tuple idxTerms = np.unravel_index(idx, self._dims)
        cdef intsize ii, cnt = len(self._content)
        cdef np.ndarray arrRes

        arrRes = self._content[0].getCols(idxTerms[0]).astype(self.dtype)
        for ii in range(1, cnt):
            arrRes = np.kron(arrRes, self._content[ii].getCols(idxTerms[ii]))

        return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        cdef tuple idxTerms = np.unravel_index(idx, self._dims)
        cdef intsize ii, cnt = len(self._content)
        cdef np.ndarray arrRes

        arrRes = self._content[0].getRows(idxTerms[0]).astype(self.dtype)
        for ii in range(1, cnt):
            arrRes = np.kron(arrRes, self._content[ii].getRows(idxTerms[ii]))

        return arrRes

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.prod(np.array(
            [term._getLargestEV(
                maxSteps, relEps, eps, alwaysReturn).astype(np.float64)
             for term in self._content]))

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.prod(np.array(
            [term._getLargestSV(
                maxSteps, relEps, eps, alwaysReturn).astype(np.float64)
             for term in self._content]))

    cpdef Matrix _getNormalized(self):
        # redo normalization for product terms, which were normalized for a
        # different data type. Otherwide the best possible accuracy cannot be
        # achieved if a wide-type kronecker product containing matrices with
        # narrow data types.
        cdef list terms = list(self._content)
        cdef Matrix term
        cdef intsize tt

        if (self.dtypeNum == TYPE_COMPLEX128 or self.dtypeNum == TYPE_FLOAT64):
            for tt in range(len(terms)):
                term = terms[tt]
                if not (term.dtypeNum == TYPE_COMPLEX128 or
                        term.dtypeNum == TYPE_FLOAT64):
                    terms[tt] = Product(term, typeExpansion=np.float64)

        for tt in range(len(terms)):
            terms[tt] = terms[tt].normalized
        return Kron(*terms)

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        cdef tuple idxTermsN = np.unravel_index(idxN, self._dims)
        cdef tuple idxTermsM = np.unravel_index(idxM, self._dims)
        cdef intsize ii, cnt = len(self._content)

        numResult = self._content[0][idxTermsN[0], idxTermsM[0]].astype(
            self.dtype)

        for ii in range(1, cnt):
            numResult *= self._content[ii][idxTermsN[ii], idxTermsM[ii]]

        return numResult

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef intsize N = len(self._content)
        return (2 * N * self.numN, 2 * N * self.numN)

    cpdef _exploreNestedProfiles(self):
        '''
        Explore the runtime properties of all nested fastmat matrices. Use ane
        iterator on self._content by default to sum the profile properties of
        all nested classes of meta-classes by default. basic-classes either
        have an empty tuple for _content or need to overwrite this method.
        '''
        cdef Matrix item
        cdef intsize scale
        cdef bint bypass
        for item in self:
            scale = self.numN / item.numN
            bypass = (item._bypassAllow and
                      (item._array is not None or item._bypassAutoArray))
            profileUpdate(&(self._profileForward), scale, bypass,
                          &(item._profileForward), &(item._profileBypassFwd))
            profileUpdate(&(self._profileBackward), scale, bypass,
                          &(item._profileBackward), &(item._profileBypassBwd))

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''

        # detect if we have an array of signals (assume the vector to
        # be two-dimensional)
        cdef intsize numSize = arrX.shape[0]
        cdef intsize numVecs = arrX.shape[1]
        cdef intsize ii, termCnt = len(self._content)
        cdef Matrix term
        cdef np.ndarray arrData = arrX

        # keep track of the identity compositions' head identity size
        cdef intsize headIN = 1

        for ii in range(termCnt):
            term = self._content[ii]

            # increase size of identity composition for this term
            headIN *= term.numN

            # reshape to match composition size
            arrData = _arrReshape(
                arrData,
                2, headIN, numVecs * self.numN / headIN,
                np.NPY_CORDER)
            # reshape to match term size
            arrData = _arrReshape(
                arrData,
                2, term.numN, numVecs * self.numN / term.numN,
                np.NPY_FORTRANORDER)

            # apply transform and reshape back to current comp. size
            arrData = _arrReshape(
                term.forward(arrData),
                2, headIN, numVecs * self.numN / headIN,
                np.NPY_FORTRANORDER)

        # reshape to matrix output dimensions
        return _arrReshape(arrData, 2, self.numN, numVecs, np.NPY_CORDER)

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''

        # detect if we have an array of signals (assume the vector to
        # be two-dimensional)
        cdef intsize numSize = arrX.shape[0]
        cdef intsize numVecs = arrX.shape[1]
        cdef intsize ii, termCnt = len(self._content)
        cdef Matrix term
        cdef np.ndarray arrData = arrX

        # keep track of the identity compositions' head identity size
        cdef intsize headIN = 1

        for ii in range(termCnt):
            term = self._content[ii]

            # increase size of identity composition for this term
            headIN *= term.numN

            # reshape to match composition size
            arrData = _arrReshape(
                arrData,
                2, headIN, numVecs * self.numN / headIN,
                np.NPY_CORDER)
            # reshape to match term size
            arrData = _arrReshape(
                arrData,
                2, term.numN, numVecs * self.numN / term.numN,
                np.NPY_FORTRANORDER)

            # apply transform and reshape back to current comp. size
            arrData = _arrReshape(
                term.backward(arrData),
                2, headIN, numVecs * self.numN / headIN,
                np.NPY_FORTRANORDER)

        # reshape to matrix output dimensions
        return _arrReshape(arrData, 2, self.numN, numVecs, np.NPY_CORDER)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes
        cdef int ff, factorCount = len(self._content)

        arrRes = self._content[0].reference().astype(
            np.promote_types(self.dtype, np.float64))
        for ff in range(1, factorCount):
            arrRes = np.kron(arrRes, self._content[ff].reference())

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 5 * 4 * 3,
                TEST.NUM_M      : TEST.NUM_N,
                'mType1'        : TEST.Permutation(TEST.ALLTYPES),
                'mType2'        : TEST.Permutation(TEST.ALLTYPES),
                'arr1'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : (5, 5)
                }),
                'arr2'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType2',
                    TEST.SHAPE  : (4, 4)
                }),
                'arr3'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : (3, 3)
                }),
                TEST.INITARGS: (lambda param : [Matrix(param['arr1']()),
                                                Matrix(param['arr2']()),
                                                Matrix(param['arr3']())]),
                TEST.OBJECT     : Kron,
                TEST.NAMINGARGS : dynFormat("%so%so%s", 'arr1', 'arr2', 'arr3'),
                TEST.TOL_POWER  : 4.
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH, arrTestDist
        from .Fourier import Fourier
        from .Diag import Diag
        from .Matrix import Matrix
        from .Eye import Eye
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Kron(
                    Fourier(2 * c),
                    Diag(np.random.uniform(2, 3, (2 * c))),
                    Matrix(arrTestDist((2 * c, 2 * c), dtype=np.complex)))),
                BENCH.FUNC_SIZE : (lambda c: 8 * c ** 3)
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Kron(*([Eye(2)] * c))),
                BENCH.FUNC_SIZE : (lambda c: (2) ** c)
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: Kron(
                    Fourier(2 * c),
                    Diag(np.random.uniform(2, 3, (2 * c)).astype(dt)),
                    Matrix(arrTestDist((2 * c, 2 * c), dt)))),
                BENCH.FUNC_SIZE : (lambda c: 8 * c ** 3)
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Kronecker Product (\texttt{fastmat.Kron})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For matrices $\bm A_i \in \C^{n_i \times n_i}$ for $i = 1,\dots,k$ the
Kronecker product
\[\bm A_1 \otimes \bm A_2 \otimes \dots \otimes \bm A_k\]
can be defined recursively because of associativity from the Kronecker product
of $\bm A \in \C^{n \times m}$ and $\bm B \in \C^{r \times s}$ defined as
\[\bm A \otimes \bm B =
\left(\begin{array}{ccc}
    a_{11} \bm B    & \dots     & a_{1m} \bm B  \\
    \vdots          & \ddots    & \vdots        \\
    a_{n1} \bm B    & \dots     & a_{nm} \bm B
\end{array}\right).\]
We make use of a decomposition into a standard matrix product to speed up the
matrix-vector multiplication which is introduced in
\cite{kron_fernandes1998automata_networks}. This then yields multiple benefits:
\begin{itemize}
\item It already brings down the complexity of the forward and backward
    projection if the factors $\bm A_i$ have no fast transformations.
\item It is not necessary to compute the matrix representation of the product,
    which saves \textbf{a lot} of memory.
\item When fast transforms of the factors are available the calculations can be
    sped up further.
\end{itemize}""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the factors',
                            'C = fm.Circulant(x_C)',
                            'H = fm.Hadamard(n)',
                            '',
                            '# define the Kronecker',
                            '# product',
                            'P = fm.Kron(C.H, H)',
                            caption=r"""
Assume we have a circulant matrix $\bm C$ with first column $\bm x_c$ and a
Hadamard matrix $\bm{\mathcal{H}}_n$ of order $n$. Then we define
    \[\bm P = \bm C^H \otimes \bm H_n.\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All but the overhead benchmarks were performed on a matrix
$\bm K = \bm \Fs_{2k} \otimes \bm D_{2k} \otimes \bm M_{2k}$, where $\bm \Fs$
is a Fourier Matrix, $\bm D$ is diagonal and $\bm M$ is unstructured and
complex valued; so $n = 8 k^3$ for $k \in \N$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(doc=r"""
$\bm K = \prod_k \bm I_k$;
so $n=k^4$ for $k \in \N$"""),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            ),
            DOC.BIBLIO(
                kron_fernandes1998automata_networks=DOC.BIBITEM(
                    r"""
Fernandes, Paulo and Plateau, Brigitte and Stewart, William J.,""",
                    r"""
Efficient Descriptor-Vector Multiplications in Stochastic Automata Networks""",
                    r'Journal of the ACM, New York, Volume 45, 1998')
            )
        )



# -*- coding: utf-8 -*-
'''
  demo/matrixSAFT.py
 -------------------------------------------------- part of the fastmat demos

  Demonstration on how to use fastMat for SAFT.


  Author      : wcw
  Introduced  :
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import sys
import timeit

# import numpy functionality
import numpy as np
import scipy as sp

################################################## import modules
try:
    import fastmat
except ImportError:
    sys.path.append('..')
    import fastmat
from fastmat.core.resource import getMemoryFootprint

###############################################################################
# SAFT demo
###############################################################################

print("fastmat demo: Synthetic Aperture Focussing Technique (SAFT)")
print("-----------------------------------------------------------")

# x - Parameter für die Matrix (Unterscheidung)
# a - Rekovektor
# b -  Messvektor
# s_x - Spezifisches X
# j - Zeilenindex Matrix - Zeitrichtung (numN)
# dx - Schrittweite in Breite (M) (numD)
# dz - Schrittweite in Tiefe (N) (numS)

numN        = 2040
numK        = 50
numD        = 1000 if numN == 2040 else numN / 2
dX          = 0.5
dZ          = 0.034

for key, value in {
    'Samples in time dimension (numN)': numN,
    'Samples in spatial dimension (numD)': numD,
    'Summation window width (numK)': numK}.items():
    print("%40s = %g" %(key, value))



print("\nGenerating masks and SAFT matrices:")



##################################################  Generate SAFT aperture masks
print(" %-20s : %s" %("'matSparse'",
                      "SAFT aperture masks as scipy-sparse matrices"))

# create an array to store index mapping
# (row-index [:, 0] to col-index[:, 1], data is [:, 2])
arrI = np.zeros((numN, 3), dtype=np.int32)
arrI[:, 0] = np.arange(numN)
arrI[:, 2] = np.ones(numN, dtype=np.int32)

sizeFull = numN * numD
sizeItem = numN
cntBlocks = sizeFull // sizeItem

# Generate sparse matrices: iterate over numK
matSparse = []
for kk in range(0, numK):
    # determine row-to-col index mapping for kk
    arrI[:, 1] = np.sqrt((kk * dX) ** 2 + (arrI[:, 0] * dZ) ** 2) / dZ + 0.5

    # select all entries within matrix range
    selection = np.where((arrI[:, 1] < sizeItem) * (arrI[:, 1] >= 0))[0]
    mapping = arrI[selection].astype(np.int16)

    # stop when submatrices are empty
    if selection.size < 1:
        break

    matSparse.append(
        sp.sparse.coo_matrix(
                    (mapping[:, 2], (mapping[:, 0], mapping[:, 1])),
                    shape=(sizeItem, sizeItem), dtype=np.int8))
fastmatSparse = [fastmat.Sparse(mat.tocsc()) for mat in matSparse]
cntK = len(matSparse)




###############################################################################
###  Generate SAFT matrix using fastmat Kronecker approach
###############################################################################

################################################## multi-diag-eye fastmat class
# define a fastmat class for a special type of eye matrix
class SparseEye(fastmat.Matrix):
    def __init__(self, numN, offset):
        self._offset = offset
        self._start = offset
        self._stop = numN - offset
        self._initProperties(numN, numN, np.int8)

    def _forward(self, arrX):
        arrRes = np.zeros((self.numN, arrX.shape[1]), dtype=arrX.dtype)
        arrRes[:self._stop, :] = arrX[self._start:, :]
        if self._offset > 0:
            arrRes[self._start:, :] += arrX[:self._stop, :]
        return arrRes

    def _backward(self, arrX):
        return self._forward(arrX)

################################################## generate actual matrix
print(" %-20s : %s" %("'matSaftKron'",
                      "using fastmat built-in classes"))
matSaftKron = fastmat.Sum(
    *[fastmat.Kron(
            fastmat.Eye(numD) if kk == 0 else SparseEye(numD, kk),
            fastmat.Sparse(mat.tocsc()))
      for kk, mat in enumerate(matSparse)])
print(" " * 4 + repr(matSaftKron))




###############################################################################
###  Generate SAFT matrix with a user-defined fastmat clas
###############################################################################

################################################## user-defined SAFT class
# define a fastmat class representing SAFT
class SaftClass(fastmat.Matrix):
    def __init__(self, numN, sizeItem, *items):
        self._items = items
        self._sizeItem = sizeItem
        self._numBlocks = numN // sizeItem
        self._numItems = len(items)
        self._initProperties(numN, numN, np.int8)

    def _core(self, arrX, backward):
        numI = self._sizeItem
        numB = self._numBlocks

        arrRes = np.empty((self.numN, arrX.shape[1]), dtype=arrX.dtype)
        viewIn  = [arrX[mm * numI:(mm + 1) * numI, :] for mm in range(numB)]
        viewOut = [arrRes[nn * numI:(nn + 1) * numI, :] for nn in range(numB)]

        for kk in range(self._numItems):
            item = self._items[kk]
            itemDot = item.vdot if backward else item.dot
            for nn in range(numB):
                if kk == 0:
                    arrArg = viewIn[nn]
                    viewOut[nn][:] = itemDot(arrArg)
                else:
                    kkl, kkr = nn - kk, nn + kk
                    arrArg = (viewIn[kkr] if kkl < 0
                              else (viewIn[kkl] if kkr >= numB
                                    else viewIn[kkl] + viewIn[kkr]))
                    viewOut[nn][:] += itemDot(arrArg)

        return arrRes

    def _forward(self, arrX):
        return self._core(arrX, False)

    def _backward(self, arrX):
        return self._core(arrX, True)

################################################## generate actual matrix
print(" %-20s : %s" %("'matSaftClass'",
                      "user-defined fastmat class"))
matSaftClass = SaftClass(sizeFull, sizeItem, *matSparse)
print(" " * 4 + repr(matSaftClass))

###############################################################################
# fastmat matrix class for efficient mask matrices
###############################################################################

################################################## import cython-optimized core
import pyximport; pyximport.install()
from matrixSAFTmask import SaftMaskCore

################################################## user-defined SAFT class
class SaftMaskClass(fastmat.Matrix):
    def __init__(self, numBlocks, sizeItem, *masks):
        if any((mask.ndim != 2 or mask.shape[1] != 2) for mask in masks):
            raise ValueError("Masks of SaftMask must be of shape Nx2!")
        self._masks = tuple(mask.astype(np.uint16, copy=True, order='C')
                            for mask in masks)
        self._sizeItem = sizeItem
        self._numBlocks = numBlocks
        self._numMasks = len(masks)
        numN = numBlocks * self._sizeItem
        self._initProperties(numN, numN, np.int8)

    def _forward(self, arrX):
        return SaftMaskCore(arrX, self._sizeItem, self._numBlocks, self._masks,
                            False)

    def _backward(self, arrX):
        return SaftMaskCore(arrX, self._sizeItem, self._numBlocks, self._masks,
                            True)

################################################## Use special fastmat class
print(" %-20s : %s" %("'matSaftMaskClass'",
                      "user-defined fastmat class, cython-optimized"))
matSaftMaskClass = SaftMaskClass(numD, sizeItem,
                                 *(tuple(np.vstack((mat.row, mat.col)).T
                                         for mat in matSparse)))
print(" " * 4 + repr(matSaftMaskClass))




###############################################################################
# Generate a sparse reference matrix with scipy (as reference)
###############################################################################
print(" %-20s : %s" %("'matSaftReference'",
                      "using scipy only"))

# determine element count upfront to initialize storage memory efficiently:
numElements = 0
for kk, mat in enumerate(matSparse):
    numElements += (1 if kk == 0 else 2) * (cntBlocks - kk) * mat.nnz
arrCol = np.empty((numElements), dtype=np.int32)
arrRow = np.empty((numElements), dtype=np.int32)
arrData = np.empty((numElements), dtype=np.int8)

# now fill that array
ii = 0
for nn in range(0, cntBlocks):
    offsetRow = nn * sizeItem
    for mm in range(0, cntBlocks):
        offsetCol = mm * sizeItem
        kk = abs(nn - mm)
        if kk >= cntK:
            continue
        mat = matSparse[kk]
        arrCol[ii:ii + mat.nnz] = mat.col + offsetCol
        arrRow[ii:ii + mat.nnz] = mat.row + offsetRow
        arrData[ii:ii + mat.nnz] = mat.data
        ii += mat.nnz
    sys.stdout.write("  [converting row %d/%d]\r" %(nn, cntBlocks))
    sys.stdout.flush()
matSaftReference = sp.sparse.coo_matrix((arrData, (arrRow, arrCol)),
    shape=(sizeFull, sizeFull), dtype=np.int8)
print(" " * 4 + repr(matSaftReference))




###############################################################################
# Generate data vectors to apply the matrices to
###############################################################################




###############################################################################
# Benchmark performance and accuracy of matrices
###############################################################################

# define a sample data vector
v = np.arange(sizeFull).reshape((sizeFull, 1)).astype(np.int64)

print("\nChecking accuracy of the various methods:")
print(" - metric used: distance norm of transform output for a given vector")
vReference = matSaftReference.dot(v)
for name, matrix in sorted({
    "matSaftKron"       : matSaftKron,
    "matSaftClass"      : matSaftClass,
    "matSaftMaskClass"  : matSaftMaskClass}.items()):
    print("%40s : %12.4e" %("%s(v) vs. reference(v)" %(name),
                            np.linalg.norm(matrix * v - vReference)))


print("\nBenchmarking forward transform performance of matrices:")
def timing(title, fun, reps):
    print("%40s : %10.4g s"% (title, timeit.timeit(fun, number=reps) / reps))

for title, fun in sorted({
    "matSaftKron"       : lambda: matSaftKron * v,
    "matSaftClass"      : lambda: matSaftClass * v,
    "matSaftMaskClass"  : lambda: matSaftMaskClass * v,
    "matSaftReference"  : lambda: matSaftReference.dot(v)}.items()):
    timing(title, fun, 1)


print("\nRAM usage for the various variants:")
for size, title in sorted({
    getMemoryFootprint(matSaftKron)         : "matSaftKron",
    getMemoryFootprint(matSaftClass)        : "matSaftClass",
    getMemoryFootprint(matSaftMaskClass)    : "matSaftMaskClass",
    getMemoryFootprint(matSaftReference)    : "matSaftReference",
    getMemoryFootprint(matSparse)           : "matSparse"}.items()):
    print("%40s : %10d kBytes" %(title, size / 1000))



# -*- coding: utf-8 -*-
'''
  fastmat/Toeplitz.py
 -------------------------------------------------- part of the fastmat package

  Toeplitz matrix. The first column is vecC1, the first row beginning at the
  second element is vecC2. We simply use a special circulant matrix to simulate
  a Toeplitz matrix. Toeplitz matrices can be embedded into larger circulant
  matrices. These are diagonal in the fourier domain and hence their multipli-
  cation with a vector can be carried out efficiently.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .core.types cimport *
from .core.cmath cimport *
from .Matrix cimport Matrix
from .Partial cimport Partial
from .Product cimport Product
from .Fourier cimport Fourier
from .Diag cimport Diag


################################################################################
################################################## class Toeplitz
cdef class Toeplitz(Partial):

    ############################################## class properties
    # vecC - Property (read-only)
    # Return the column-defining vector of Toeplitz matrix.
    property vecC:
        def __get__(self):
            return self._vecC

    # vecR - Property (read-only)
    # Return the row-defining vector of Toeplitz matrix.
    property vecR:
        def __get__(self):
            return self._vecR

    ############################################## class methods
    def __init__(self, vecC, vecR, **options):
        '''
        Initialize Toeplitz Matrix instance.

        The toeplitz matrix is embedded into a circulant matrix. See class
        Circulant for further details.

        Toeplitz([ c C C C C ], [ r R R R ]) represents this matrix:
          [ c r R R R ]
          [ C c r R R ]
          [ C C c r R ]
          [ C C C c r ]
          [ C C C C c ]

        The Circulant it is derived from has the following generator:
          [ c C C C C R R R r ] [9 x 9]

        ...and may be zero-padded to a [16 x 16] Matrix for efficiency:
          [ c C C C C 0 0 0 0 0 0 0 R R R r]

        Valid options
        -------------
         'pad'=[false TRUE]
            perform zero-padding of Circulant for efficiency

        All options specified will also be passed on to the geneation of the
        underlying Product instance initialization.
        '''

        # save generating vectors. Matrix sizes will be set by Product
        dataType = np.promote_types(vecC.dtype, vecR.dtype)
        self._vecC = vecC.astype(dataType)
        self._vecR = vecR.astype(dataType)

        # evaluate options passed to class
        cdef bint optimize = options.get('optimize', True)
        cdef int maxStage = options.get('maxStage', 4)

        # perform padding (if enabled) and generate vector
        cdef intsize size = len(vecC) + len(vecR)
        cdef intsize vecSize = size

        # determine if zero-padding of the convolution to achieve a better FFT
        # size is beneficial or not
        if optimize:
            vecSize = _findOptimalFFTSize(size, maxStage)

            assert vecSize >= size

            if _getFFTComplexity(size) <= _getFFTComplexity(vecSize):
                vecSize = size

        if vecSize > size:
            # zero-padding pays off, so do it!
            vec = np.concatenate([self._vecC,
                                  np.zeros((vecSize - size,),
                                           dtype=dataType),
                                  np.flipud(self._vecR)])
        else:
            vec = np.concatenate([self._vecC, np.flipud(self._vecR)])

        # Describe as circulant matrix with product of data and vector
        # in fourier domain. Both fourier matrices cause scaling of the
        # data vector by N, which will be compensated in Diag().

        # Create inner product
        cdef Fourier FN = Fourier(vecSize)
        cdef Product P = Product(FN.H, Diag(np.fft.fft(vec) / vecSize),
                                 FN, **options)

        # initialize Partial of Product
        kwargs = {}
        if size != len(self._vecC):
            kwargs['N'] = np.arange(len(self._vecC))

        if size != len(self._vecR) + 1:
            kwargs['M'] = np.arange(len(self._vecR) + 1)

        super(Toeplitz, self).__init__(P, **kwargs)

        # Currently Fourier matrices bloat everything up to complex double
        # precision, therefore make sure vecC and vecR matches the precision of
        # the matrix itself
        if self.dtype != self._vecC.dtype:
            self._vecC = self._vecC.astype(self.dtype)

        if self.dtype != self._vecR.dtype:
            self._vecR = self._vecR.astype(self.dtype)

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        cdef intsize N = self.numN
        cdef np.ndarray arrRes

        if idx == 0:
            return self._vecC
        elif idx >= N:
            # double slicing needed, otherwise fail when M = N + 1
            return self._vecR[idx - N:idx][::-1]
        else:
            arrRes = _arrEmpty(1, N, 0, self._info.dtype[0].typeNum)
            arrRes[:idx] = self._vecR[idx - 1::-1]
            arrRes[idx:] = self._vecC[:N - idx]
            return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        cdef intsize M = self.numM
        cdef np.ndarray arrRes

        if idx >= M - 1:
            # double slicing needed, otherwise fail when N = M + 1
            return self._vecC[idx - M + 1:idx + 1][::-1]
        else:
            arrRes = _arrEmpty(1, M, 0, self._info.dtype[0].typeNum)
            arrRes[:idx + 1] = self._vecC[idx::-1]
            arrRes[idx + 1:] = self._vecR[:M - 1 - idx]
            return arrRes

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        cdef intsize distance = idxN - idxM
        return (self._vecR[-distance - 1] if distance < 0
                else self._vecC[distance])

    cpdef np.ndarray _getArray(self):
        '''
        Return an explicit representation of the matrix as numpy-array.
        '''
        return self._reference()

    cpdef Matrix _getNormalized(self):
        # NOTE: This method suffers accuracy losses when elements with lower
        # indices are large in magnitude compared to ones with higher index!
        cdef intsize ii, N = self.numN, M = self.numM
        cdef intsize iiMax = (N + 1) if M > N else M

        # fill in a placeholder array
        cdef np.ndarray arrNorms = _arrZero(1, M, 1, np.NPY_FLOAT64)

        # compute the absolute value of the squared elements in the defining
        # vectors
        cdef np.ndarray vecCSqr = np.square(np.abs(self._vecC))
        cdef np.ndarray vecRSqr = np.square(np.abs(self._vecR))

        # the first column is easy
        arrNorms[0] = vecCSqr.sum()

        # then follow the iterative approach. Every subsequent column features
        # one more element of vecR and one less of vecC. Continue until vecC is
        # eaten up completely
        for ii in range(1, iiMax):
            arrNorms[ii] = arrNorms[ii - 1] + vecRSqr[ii - 1] - vecCSqr[N - ii]

        # then (as vecC is eaten up), proceed with rolling the remaining
        # elements of vecR until they are represented fully in arrNorms
        for ii in range(iiMax, M):
            arrNorms[ii] = (arrNorms[ii - 1] +
                            vecRSqr[ii - 1] - vecRSqr[ii - iiMax])

        return self * Diag(1. / np.sqrt(arrNorms))

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (0., 0.)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        _reference borrowing from Partial is too slow. Therefore,
        construct a reference directly from the vectors.
        '''
        cdef intsize ii, N = self.numN, M = self.numM
        cdef np.ndarray arrRes = np.empty((N, M), dtype=self.dtype)

        # put columns in lower-triangular part of matrix
        for ii in range(0, min(N, M)):
            arrRes[ii:N, ii] = self._vecC[0:(N - ii)]

        # put rows in upper-triangular part of matrix
        for ii in range(0, min(N, M - 1)):
            arrRes[ii, (ii + 1):M] = self._vecR[0:(M - ii - 1)]

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # 35 is just any number that causes no padding
                # 41 is the first size for which bluestein is faster
                TEST.NUM_N      : TEST.Permutation([5, 41]),
                'num_M'         : TEST.Permutation([4, 6]),
                TEST.NUM_M      : (lambda param: param['num_M'] + 1),
                'mTypeH'        : TEST.Permutation(TEST.ALLTYPES),
                'mTypeV'        : TEST.Permutation(TEST.FEWTYPES),
                'optimize'      : True,
                'vecH'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeH',
                    TEST.SHAPE  : (TEST.NUM_N, )
                }),
                'vecV'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeV',
                    TEST.SHAPE  : ('num_M', )
                }),
                TEST.INITARGS   : (lambda param : [param['vecH'](),
                                                   param['vecV']()]),
                TEST.INITKWARGS : {'optimize' : 'optimize'},
                TEST.OBJECT     : Toeplitz,
                TEST.NAMINGARGS : dynFormat("%s,%s,optimize=%s",
                                            'vecH', 'vecV', str('optimize')),
                TEST.TOL_POWER  : 2
            },
            TEST.CLASS: {
                # perform thorough testing of slicing during array construction
                # therefore, aside the symmetric shape case also test shapes
                # that differ by +1/-1 and +x/-x in row and col size
                TEST.NUM_N      : 4,
                'num_M'         : TEST.Permutation([2, 3, 4, 5, 6]),
            },
            TEST.TRANSFORMS: {
                # test differences in padding only for the transforms
                'padding'       : TEST.Permutation([True, False]),
            }
        }

    def _getBenchmark(self):
        from .inspect import BENCH, arrTestDist
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Toeplitz(
                    arrTestDist((c, ), dtype=np.float),
                    arrTestDist((c - 1, ), dtype=np.float)))
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Toeplitz(
                    arrTestDist((2 ** c, ), dtype=np.float),
                    arrTestDist((2 ** c - 1, ), dtype=np.float)))
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, datatype: Toeplitz(
                    arrTestDist((2 ** c, ), dtype=datatype),
                    arrTestDist((2 ** c - 1, ), dtype=datatype)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Toeplitz Matrix (\texttt{fastmat.Toeplitz})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
A Toeplitz matrix $\bm T \in \C^{n \times m}$ realizes the mapping
\[\bm x \mapsto \bm T \cdot \bm x,\]
where $\bm x \in C^n$ and
\[\bm T = \left(\begin{array}{cccc}
    t_1 & t_{-1} & \dots & t_{-(m-1)} \\
    t_2 & t_1 & \ddots & t_{-(n-2)} \\
    \vdots & \vdots & \ddots & \vdots \\
    t_n & t_{n-1} & \dots & t_1
\end{array}\right).\]
This means that a Toeplitz matrix is uniquely defined by the $n + m - 1$ values
that are on the diagonals.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# define the parameters',
                            'd1 = np.array([1,0,3,6])',
                            'd2 = np.array([5,7,9])',
                            '',
                            '# construct the transform',
                            'T = fm.Toeplitz(d1,d2)',
                            caption=r"""
This yields
\[\bm d_1 = (1,0,3,6)^T\]
\[\bm d_2 = (5,7,9)^T\]
\[\bm T = \left(\begin{array}{cccc}
    1 & 5 & 7 & 9 \\
    0 & 1 & 5 & 7 \\
    3 & 0 & 1 & 5 \\
    6 & 3 & 0 & 1
\end{array}\right)\]"""),
                r"""
Since the multiplication with a Toeplitz matrix makes use of the FFT, it can
be very slow, if the sum of the dimensions of $\bm d_1$ and $\bm d_2$ are far
away from a power of $2$, $3$ or $4$. This can be alleviated if one applies
smart zeropadding during the transformation.
This can be activated as follows.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# define the parameters',
                            'd1 = np.array([1,0,3,6])',
                            'd2 = np.array([5,7,9])',
                            '',
                            '# construct the transform',
                            "T = fm.Toeplitz(d1,d2,pad='true')",
                            caption=r"""
This yields the same matrix and transformation as above, but it might be faster
depending on the dimensions involved in the problem."""),
                r"""
This class depends on \texttt{Fourier}, \texttt{Diag}, \texttt{Product} and
\texttt{Partial}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm T \in \R^{n \times n}$ with generating entries drawn from a standard
Gaussian distribution""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/Diag.pyx
 -------------------------------------------------- part of the fastmat package

  Diagonal matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .Eye cimport Eye
from .core.cmath cimport _conjugate, _multiply, _arrZero
from .core.types cimport *


################################################################################
################################################## class Diag
cdef class Diag(Matrix):

    ############################################## class properties
    # vecD - Property (read-only)
    # Return the matrix-defining vector of diagonal entries.
    property vecD:
        def __get__(self):
            return self._vecD

    ############################################## class methods
    def __init__(self, vecD):
        '''Initialize Matrix instance with a list of child matrices'''
        # numN is size of matrix (and of diagonal vector)
        numN = len(vecD)

        # store diagonal entry vector as copy of vecD and complain if
        # dimension does not match
        self._vecD = np.atleast_1d(np.squeeze(np.copy(vecD)))
        if self._vecD.ndim != 1:
            raise ValueError(
                "Diag: Definition vector must have exactly one dimension.")

        # set properties of matrix
        self._initProperties(
            numN, numN, self._vecD.dtype,
            cythonCall=True,
            forceInputAlignment=True,
            fortranStyle=True
        )

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        cdef np.ndarray arrRes

        arrRes = _arrZero(1, self.numN, 1, self._info.dtype[0].typeNum)
        arrRes[idx] = self._vecD[idx]

        return arrRes

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.abs(self._vecD).max().astype(np.float64)

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.abs(self._vecD).max().astype(np.float64)

    cpdef np.ndarray _getRow(self, intsize idx):
        return self._getCol(idx)

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._vecD[idxN] if idxN == idxM else self.dtype(0)

    cpdef Matrix _getGram(self):
        return Diag(np.abs(self._vecD) ** 2)

    cpdef Matrix _getNormalized(self):
        if self._info.dtype.isComplex:
            return Diag((self._vecD / abs(self._vecD)).astype(self.dtype))
        else:
            return Diag(np.sign(self._vecD).astype(self.dtype))

    cpdef Matrix _getT(self):
        return self

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (2. * self.numN, 3. * self.numN)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        ''' Calculate the forward transform of this matrix.'''
        _multiply(
            arrX, self._vecD, arrRes,
            typeX, self._info.dtype[0].fusedType, typeRes)

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        ''' Calculate the backward transform of this matrix.'''
        _multiply(
            arrX, _conjugate(self._vecD), arrRes,
            typeX, self._info.dtype[0].fusedType, typeRes)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef intsize ii, N = self.numN
        cdef np.ndarray d = self.vecD

        return np.diag(d)

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 35,
                TEST.NUM_M      : TEST.NUM_N,
                'mTypeD'        : TEST.Permutation(TEST.ALLTYPES),
                TEST.PARAMALIGN : TEST.Permutation(TEST.ALLALIGNMENTS),
                'vecD'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeD',
                    TEST.SHAPE  : (TEST.NUM_N, ),
                    TEST.ALIGN  : TEST.PARAMALIGN
                }),
                TEST.INITARGS   : (lambda param: [param['vecD']()]),
                TEST.OBJECT     : Diag,
                TEST.NAMINGARGS : dynFormat("%s", 'vecD')
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Diag(np.random.uniform(2, 3, c)))
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c:
                                   Diag(np.random.uniform(2, 3, 2 ** c)))
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: Diag(
                    np.random.uniform(2, 3, 2 ** c).astype(dt)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Diagonal Matrix (\texttt{fastmat.Diagonal})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
\[x \mapsto \bm{\mathrm{diag}}(d_1,\dots,d_n) \cdot \bm x\]
A diagonal matrix is uniquely defined by the entries of its diagonal.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# build the parameters',
                            'n = 4',
                            'd = np.array([1, 0, 3, 6])',
                            '',
                            '# construct the matrix',
                            'D = fm.Diagonal(d)',
                            caption=r"""
This yields
\[\bm d = (1, 0, 3, 6)^T\]
\[\bm D = \left(\begin{array}{cccc}
    1 & & & \\
    & 0 & & \\
    & & 3 & \\
    & & & 6
\end{array}\right)\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm D \in \R^{n \times n}$ with $n \in \N$ and diagonal entries drawn from a
uniform distribution on $[2,3]$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/inspect/common.py
 -------------------------------------------------- part of the fastmat package

  Implementations of common quality management structures and routines.


  Author      : wcw
  Introduced  : 2017-07-19
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import copy
import inspect
import itertools
import os
import re
import six
import struct
import numpy as np
from scipy import sparse
from platform import system as pfSystem

try:
    from itertools import izip
except ImportError:         # python 3.x
    izip = zip

currentOS = pfSystem()


################################################################################
################################################## CONSTANT definition classes
class ALIGNMENT(object):
    DONTCARE        = '-'
    FCONT           = 'F'
    CCONT           = 'C'
    STRIDE          = 'S'


################################################################################
################################################## Permutation funcs and classes

class AccessDict(dict):
    def __getattr__(self, key):
        if key in self:
            return self[key]
        else:
            found = [kk for kk in sorted(self.keys())
                     if kk.startswith(key)]
            return ([self[kk] for kk in found] if len(found) > 2
                    else (None if len(found) == 0
                          else self[found[0]]))

    def __repr__(self):
        return str(self.keys())


def convertToAccessDicts(level):
    for key, value in level.items():
        if isinstance(value, dict) and (value is not value):
            level[key] = convertToAccessDicts(value)

    return AccessDict(level)


################################################## class uniqueNameDict
class uniqueNameDict(dict):
    '''
    Modified dictionary: suffixes key names with integer to maintain uniqueness.
    '''

    def __setitem__(self, key, value):
        index = 1
        strKey = key
        while True:
            if strKey not in self:
                break
            index += 1
            strKey = "%s_%03d" %(key, index)

        dict.__setitem__(self, strKey, value)


################################################## class paramDict
reFormatString = re.compile(r'%\(.+\)')


class paramDict(dict):

    def __getattr__(self, key):
        # evaluate nested format-string parameters, update format results
        value, lastValue = super(paramDict, self).__getitem__(key), None

        while id(lastValue) != id(value):
            lastValue = value
            if isinstance(value, str):
                if value in self and value != key:
                    value = getattr(self, value)
                elif reFormatString.search(value):
                    value = value %self

            elif (inspect.isroutine(value) and
                  not isinstance(value, IgnoreFunc)):
                value = value(self)
                self[key] = value

        return value


################################################## class Permutation
class Permutation(list):

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           super(Permutation, self).__repr__())

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           super(Permutation, self).__str__())


################################################## class VariantPermutation
class VariantPermutation(Permutation):
    pass


################################################## class IgnoreDict
class IgnoreDict(dict):
    pass


################################################## class IgnoreFunc
class IgnoreFunc(object):

    def __init__(self, fun):
        self._fun = fun

    def __call__(self, *args, **kwargs):
        return self._fun(*args, **kwargs)


################################################## paramApplyDefaults()
def paramApplyDefaults(
    params,
    templates=None,
    templateKey=None,
    extraArgs=None
):

    # have some defaults from templates
    # first, fetch the template as defaults, then update with target,
    # then assign the whole dict to target
    result = {}
    if templates is not None:
        # 1. COMMON - section of templates (lowest-priority)
        result.update(templates.get(NAME.COMMON, {}))
        # 2. the templates-section corresponding to the templateKey
        result.update(templates.get(templateKey, {}))
        # 3. specific reference to a template by the 'template' key in params
        if NAME.TEMPLATE in params:
            result.update(templates.get(params[NAME.TEMPLATE], {}))

    # 4. actual parameters (params)
    result.update(params)

    # 5. extraArgs (which usually come from command-line) (top-priority)
    if extraArgs is not None:
        for p in extraArgs:
            for pp in list(p.split(',')):
                tokens = pp.split('=')
                if len(tokens) >= 2:
                    string = "=".join(tokens[1:])
                    try:
                        val = int(string)
                    except ValueError:
                        try:
                            val = float(string)
                        except ValueError:
                            val = string
                    result[tokens[0]] = val

    return paramDict(result)


################################################## paramPermute()
def paramPermute(dictionary, copy=True, PermutationClass=Permutation):
    '''
    Return a list of cartesian-product combination of all dictionary values
    holding a Permutation list object. If copy is True, the resulting list will
    be applied back to input dictionary copies before returning, resulting in
    a list of copies with the permutation replacements being made on the input
    dictionary. Parameter permutations must be indicated by wrapping a list in
    a Permutation class instance.
    '''
    # isolate the permutation parameters from the dictionary
    parameters = {key: list(value)
                  for key, value in dictionary.items()
                  if isinstance(value, PermutationClass)}

    # perform a cartesian product -> list of permuted instances
    permutations = [dict(izip(parameters, x))
                    for x in itertools.product(*six.itervalues(parameters))]
    if len(permutations) == 1:
        permutations = [{}]

    # apply the permutations to the input dictionary (generating copies)
    def dictCopyAndMerge(source, merge):
        result = source.copy()
        result.update(merge)
        return paramDict(result)
    return [dictCopyAndMerge(dictionary, permutation)
            for permutation in permutations] if copy else permutations


################################################## paramDereferentiate()
def paramDereferentiate(currentLevel, paramDict=None):
    '''
    Replace all text value identifyers matching a target key name with the
    key values it's pointing to allowing parameter links. Then, recurse through
    an arbitrary depth of container types (dicts, lists, tuples) found in the
    first stage, continuing dereferentiation.
    Returns the modified currentLevel container.
    '''
    # set paramDict in first level, determine iterator for currentLevel
    iterator = currentLevel.items() \
        if isinstance(currentLevel, dict) else enumerate(currentLevel)
    if paramDict is None:
        paramDict = currentLevel

    # STAGE 1: Replace all dictionary values matching a key name in paramDict
    #          with the corresponding value of that paramDict-entry. Also build
    #          a container list for stage two.
    dictIterables = {}
    for key, value in iterator:
        if isinstance(value, (list, tuple, dict)):
            dictIterables[key] = value
        else:
            if not isinstance(value, str):
                continue
            try:
                paramValue = paramDict.get(value, None)
                if (paramValue is not None and
                        not isinstance(paramValue, Permutation)):
                    currentLevel[key] = paramValue
                    if isinstance(paramValue, (list, tuple, dict)):
                        dictIterables[key] = paramValue
            except TypeError:
                continue

    # STAGE 2: Crawl the containers found in stage 1, repeating the process.
    #          Note that nested containers are copied in the process.
    #          The parameter dictionary paramDict stays the same for all levels.
    for key, iterable in dictIterables.items():
        # copy the container to allow modification of values
        newIterable = copy.copy(iterable)
        if isinstance(iterable, tuple):
            # cast the immutable tuple type to list allowing modifications, cast
            # back to tuple afterwards
            newIterable = list(newIterable)
            paramDereferentiate(newIterable, paramDict)
            newIterable = tuple(newIterable)
        else:
            paramDereferentiate(newIterable, paramDict)
        # overwrite former iterable with new copy
        currentLevel[key] = newIterable

    return currentLevel


################################################## paramEvaluate()
def paramEvaluate(currentLevel, paramDict=None):
    '''
    Evaluate all functions found in currentLevel with paramDict as argument.
    Repeat the process for nested containers.
    Returns the modified currentLevel container.
    '''
    # set paramDict in first level, determine iterator for currentLevel
    iterator = currentLevel.items() \
        if isinstance(currentLevel, dict) else enumerate(currentLevel)
    if paramDict is None:
        paramDict = currentLevel

    # STAGE 1: Evaluate the functions found in currentLevel. Also build a
    #          container list for stage two.
    dictIterables = {}
    for key, value in iterator:
        if isinstance(value, (list, tuple, dict)):
            if not isinstance(value, IgnoreDict):
                dictIterables[key] = value
        elif inspect.isroutine(value):
            currentLevel[key] = value(paramDict)

    # STAGE 2: Crawl the containers found in stage 1, repeating the process
    #          The parameter dictionary paramDict stays the same for all levels.
    for key, iterable in dictIterables.items():
        if isinstance(iterable, tuple):
            # cast the immutable tuple type to list allowing modifications, cast
            # back to tuple afterwards
            newIterable = list(currentLevel[key])
            paramEvaluate(newIterable, paramDict)
            currentLevel[key] = tuple(newIterable)
        else:
            paramEvaluate(currentLevel[key], paramDict)

    return currentLevel


################################################## mergeDicts()
# a little helper to safely merge two dictionaries
def mergeDicts(a, b):
    '''
    Merge the dictionaries a and b such that entries in b have priority and the
    input Dictionary a remains unchanged.
    '''
    c=a.copy()
    c.update(b)
    return c

################################################################################
################################################## array generators, test distr.


################################################## arrTestDist()
def arrTestDist(shape, dtype, center=0):
    def _draw(shape):
        '''
        Draw a random floating-point number from a test distribution.
        Remove the part around zero from the distribution and keep the distance
        between minimal and maximal absolute values (dynamics) relatively small
        '''
        return (np.random.uniform(2., 1., size=shape) *
                np.random.choice([-1, 1], shape))

    if np.prod(shape) < 1:
        return np.array([])

    if np.issubdtype(dtype, np.int):
        result = np.random.choice(
            [center - 2, center - 1, center + 1, center + 2], shape).astype(
                dtype)
    else:
        if np.issubdtype(dtype, np.float):
            result = _draw(shape).astype(dtype) + center
        elif np.issubdtype(dtype, np.complex):
            result = (_draw(shape) + np.real(center) +
                      1j * (_draw(shape) + np.imag(center))).astype(dtype)
        else:
            raise TypeError("arrTestDist: unsupported type %s" % (dtype))

    # increase the largest element in magnitude a little bit more to avoid
    # too close neighbours to the largest element in the distribution
    # this helps at least largestSV in Diag matrices to converge ;)
    idxMax = np.unravel_index(np.abs(result).argmax(), result.shape)
    if np.issubdtype(dtype, np.int):
        result[idxMax] += np.sign(result[idxMax])
    else:
        result[idxMax] *= 1.5

    return result

################################################## arrSparseTestDist()


def arrSparseTestDist(shape, dtype,
                      density=0.1, center=0, compactFullyOccupied=False):
    numSize = np.prod(shape)
    if compactFullyOccupied:
        # draw just enough samples randomly such that every row and column is
        # occupied with at least one element. Ignore the density parameter
        numElements = max(shape)

        # draw mm and nn coordinates from coordinate space. Modulo operation
        # wraps indices larger than the actual row or column dimension
        suppX = np.mod(
            np.random.choice(numElements, numElements, replace=False), shape[0])
        suppY = np.mod(
            np.random.choice(numElements, numElements, replace=False), shape[1])
    else:
        # draw a relative amount of samples randomly over vectorized index space
        numElements = int(numSize * density)
        supp = np.random.choice(np.arange(numSize), numElements, replace=False)
        suppX = np.mod(supp, shape[0])
        suppY = np.divide(supp, shape[0])

    # determine the actual element values distributed over the sparse array
    # from arrTestDist with a 1D-array spanning the required element count
    arrElements = arrTestDist((numElements, ), dtype, center=center)

    return sparse.coo_matrix(
        (arrElements, (suppX, suppY)), shape=shape, dtype=dtype)


################################################## arrAlign()
def arrAlign(arr, alignment=ALIGNMENT.DONTCARE):
    if alignment == ALIGNMENT.DONTCARE:
        return np.asanyarray(arr)
    elif alignment == ALIGNMENT.FCONT:
        return np.asfortranarray(arr)
    elif alignment == ALIGNMENT.CCONT:
        return np.ascontiguousarray(arr)
    elif alignment == ALIGNMENT.STRIDE:
        # define spacing between elements
        spacing = 3
        # determine maximum value
        try:
            maxValue = np.iinfo(arr.dtype).max
        except ValueError:
            try:
                maxValue = np.finfo(arr.dtype).max
            except ValueError:
                maxValue = 1.

        # generate large random array with maximized data type utilization
        arrFill = (maxValue * (np.random.rand(
            *(dim * spacing for dim in arr.shape)) - 0.5)).astype(arr.dtype)

        # fill-in the array data and return a view of the to-be-aligned array
        arrPart = arrFill[(np.s_[1::spacing], ) * arrFill.ndim]
        arrPart[:] = arr
        return arrPart
    else:
        raise ValueError("Unknown alignment identificator '%s'" %(alignment))


################################################## arrayGenerator
class ArrayGenerator(dict):

    @property
    def forwardData(self):
        if NAME.FWDATA not in self:
            # generate random array and set specific alignment style
            self[NAME.FWDATA] = arrAlign(
                arrTestDist(self[NAME.SHAPE], self[NAME.DTYPE],
                            center=self.get(NAME.CENTER, 0)),
                alignment=self.get(NAME.ALIGN, ALIGNMENT.DONTCARE))

        return self[NAME.FWDATA]

    @property
    def backwardData(self):
        if NAME.BWDATA not in self:
            # generate random array and set specific alignment style
            self[NAME.BWDATA] = arrAlign(
                arrTestDist(self[NAME.SHAPE_T], self[NAME.DTYPE],
                            center=self.get(NAME.CENTER, 0)),
                alignment=self.get(NAME.ALIGN, ALIGNMENT.DONTCARE))

        return self[NAME.BWDATA]

    def __call__(self):
        return self.forwardData

    def __str__(self):
        '''Compose a compact description of the represented array.'''
        tags = []

        # generate shape-token: check for both shape variants (fw and bw).
        # if they differ, print both as "fw/bw", otherwise print the dim only
        fwShape = self[NAME.SHAPE] if NAME.SHAPE in self else ()
        bwShape = self[NAME.SHAPE_T] if NAME.SHAPE_T in self else ()

        def printDim(fw, bw):
            if fw is None:
                return '-' if bw is None else str(bw)
            else:
                return (str(fw) if bw is None
                        else "%s/%s" %(fw, bw) if (fw != bw) else str(fw))

        strShape = 'x'.join([
            printDim(fw, bw)
            for fw, bw in six.moves.zip_longest(fwShape, bwShape)])
        if len(strShape) > 0:
            tags.append(strShape)

        # print the data type of the array
        value = self.get(NAME.DTYPE, '')
        if isinstance(value, type):
            tags.append(NAME.TYPENAME.get(value, str(value)))

        value = self.get(NAME.ALIGN, None)
        if value in NAME.ALLALIGNMENTS:
            tags.append(value)

        return str("[%s]" % (','.join(tags)))

    def __repr__(self):
        return self.__str__()


################################################################################
################################################## CONSTANT definitions

################################################## class TypeDict
class TypeDict(dict):
    def __getitem__(self, key):
        if key not in self:
            key = key.type if isinstance(key, np.dtype) else None
        return dict(self).get(key, '???')


class COLOR():
    '''Give escape format strings (color, face) a name.'''
    END    = "\033[0m"
    BOLD   = "\033[1m"
    LINE   = "\033[4m"
    RED    = "\033[91m"
    GREEN  = "\033[92m"
    YELLOW = "\033[93m"
    BLUE   = "\033[94m"
    PURPLE = "\033[95m"
    AQUA   = "\033[96m"


def fmtStr(string, color):
    '''Print a string quoted by some format specifiers.'''
    # colored output only supported with linux
    return ("%s%s%s" %(color, string, COLOR.END)
            if currentOS == 'Linux'
            else string)


def fmtGreen(string):
    '''Print string in green.'''
    return fmtStr(string, COLOR.GREEN)


def fmtRed(string):
    '''Print string in red.'''
    return fmtStr(string, COLOR.RED)


def fmtYellow(string):
    '''Print string in yellow.'''
    return fmtStr(string, COLOR.YELLOW)


def fmtBold(string):
    '''Print string in bold face.'''
    return fmtStr(string, COLOR.BOLD)


reAnsiEscape = None


def fmtEscape(string):
    '''Return a string with all ASCII escape sequences removed.'''
    global reAnsiEscape
    if reAnsiEscape is None:
        reAnsiEscape = re.compile(r'\x1b[^m]*m')

    return reAnsiEscape.sub('', string)


def dynFormat(s, *keys):
    return s.replace('%', '%%(%s)') % keys


################################################## getConsoleSize()

fallbackConsoleSize = (80, 25)
if (currentOS in ['Linux', 'Darwin']) or currentOS.startswith('CYGWIN'):
    import fcntl
    import termios
    # source: https://gist.github.com/jtriley/1108174

    def getConsoleSize():
        def ioctl_GWINSZ(fd):
            try:
                return struct.unpack(
                    'hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
            except EnvironmentError:
                return None

        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)
            except EnvironmentError:
                pass

        if not cr:
            try:
                cr = (os.environ['LINES'], os.environ['COLUMNS'])
            except (EnvironmentError, KeyError):
                cr = fallbackConsoleSize

        return int(cr[0]), int(cr[1])
elif currentOS == 'Windows':
    def getConsoleSize():
        cr = fallbackConsoleSize
        try:
            from ctypes import windll, create_string_buffer
            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (left, top, right, bottom) = struct.unpack("10x4h4x", csbi.raw)
                cr = (right - left, bottom - top)
        except (ImportError, EnvironmentError):
            pass

        return cr
else:
    def getConsoleSize():
        return fallbackConsoleSize


################################################## worker's CONSTANT classes
class NAME(object):
    DATA            = 'data'
    FWDATA          = 'dataForward'
    BWDATA          = 'dataBackward'
    DTYPE           = 'dtype'
    SHAPE           = 'shape'
    SHAPE_T         = 'shapeBackward'
    ALIGN           = 'align'
    CENTER          = 'center'
    FORMAT          = 'format'
    NAME            = 'name'

    CLASS           = 'class'
    TARGET          = 'target'
    FILENAME        = 'filename'
    CAPTION         = 'caption'
    BENCHMARK       = 'bench'
    DOCU            = 'docu'
    TEST            = 'test'
    COMMON          = 'common'
    TEMPLATE        = 'template'
    VARIANT         = 'variant'
    RESULT          = 'result'
    HEADER          = 'header'

    TYPENAME        = TypeDict({
        np.int8:       'i08', np.int32:      'i32', np.int64:      'i64',
        np.float32:    'f32', np.float64:    'f64',
        np.complex64:  'c32', np.complex128: 'c64', None:          '???'
    })

    ALLTYPES        = [np.int8, np.int32, np.int64,
                       np.float32, np.float64,
                       np.complex64, np.complex128]
    FEWTYPES        = [np.int8, np.int32, np.float32, np.complex64]

    SINGLETYPES     = [np.int32, np.float32, np.complex64]
    DOUBLETYPES     = [np.int64, np.float64, np.complex128]

    INTTYPES        = [np.int8, np.int32, np.int64]
    FLOATTYPES      = [np.float32, np.float64]
    COMPLEXTYPES    = [np.complex64, np.complex128]

    LARGETYPES      = [np.int32, np.int64, np.float32, np.float64,
                       np.complex64, np.complex128]

    ALLALIGNMENTS   = [ALIGNMENT.FCONT, ALIGNMENT.CCONT, ALIGNMENT.STRIDE]

    # repeat some of the definitions in this unit to allow compact imports
    ALIGNMENT       = ALIGNMENT
    ArrayGenerator  = ArrayGenerator
    Permutation     = Permutation
    IgnoreFunc      = IgnoreFunc


################################################################################
################################################## class Runner
class Worker(object):
    '''
    options - dictionary structure containing options for multiple targets.
        {   'nameOfTarget': {'parameter': 123,
                                'anotherParameter': 456
         }, NAME.DEFAULTS:  {'parameter': default parameter unless overwritten
                                          within the selected target
         }}

    results - output of each target's as specified in options
    '''

    cbStatus=None
    cbResult=None
    target=None

    def __init__(self, targetClass, **options):
        '''
        Setup an inspection environment on a fastmat class specified in target.
        Along the way an empty instance of target will be created and aside
        the default options, which may be specified in runnerDefaults, an
        arbitrarily named method of target may be specified to return more
        specific options when called.
        extraOptions may be specified to overwrite any parameter with highest
        priority.
        Both extraOptions and runnerDefaults must be specified as a two-level
        dictionary with the outer level specifying target names as keys and the
        inner level the actual parameter for the target.
        '''
        # the test target is a fastmat class to be instantiated in the runners
        if not inspect.isclass(targetClass):
            raise ValueError("target in init of Runner must be a class type")

        # set defaults for options
        targetOptionMethod = options.get('targetOptionMethod', None)
        runnerDefaults = options.get('runnerDefaults', {})
        extraOptions = options.get('extraOptions', {})

        self.target=targetClass.__new__(targetClass)

        # targetOptionMethod specifies a method name of target which will
        # return a dictionary with class-specific options if called. If this
        # functionality is not needed, targetOptionsMethod or runnerDefaults
        # may be left to their default values.
        targetOptionMethod=getattr(self.target, targetOptionMethod, None)
        options={name: mergeDicts(target, extraOptions)
                 for name, target in ({} if targetOptionMethod is None
                                      else targetOptionMethod()).items()}

        # determine keys for final output
        keys=[name for name in options.keys() if name != NAME.COMMON]

        # start with the defaults for the selected keys as a baseline
        common=runnerDefaults.get(NAME.COMMON, {})
        self.options={name: mergeDicts(common, runnerDefaults.get(name, {}))
                      for name in keys}

        # a NAME.TEMPLATE parameter in a target will cause the specified
        # target to be extend by the default options of another target.
        # Priority order: extraOptions > options > defaults > COMMON
        for name, target in self.options.items():
            template=(options[name][NAME.TEMPLATE]
                      if name in options and NAME.TEMPLATE in options[name]
                      else (runnerDefaults[name][NAME.TEMPLATE]
                            if (name in runnerDefaults and
                                NAME.TEMPLATE in runnerDefaults[name])
                            else None))

            if template is not None:
                target.update(
                    runnerDefaults.get(template, {}))

        # now add our specific options from options and extraoptions
        common=options.get(NAME.COMMON, {})
        self.options={name: mergeDicts(self.options.get(name, {}),
                                       mergeDicts(common, options[name]))
                      for name in keys}

        # finally, tag each target with its target name so that parameter links
        # may address it over the key [NAME.TARGET]. Also, write down the name
        # of the class in [NAME.CLASS]
        for name, target in self.options.items():
            target[NAME.TARGET]=name
            target[NAME.CLASS]=targetClass.__name__

        # initialize output
        self.results=AccessDict({})

    def emitStatus(self, *args):
        if self.cbStatus is not None:
            self.cbStatus(*args)

    def emitResult(self, *args):
        if self.cbResult is not None:
            self.cbResult(*args)

    def run(self, *targetNames, **extraArgs):
        '''
        Execute all selected targets by a list of targetNames.
        If *targetNames is empty all targets will be run.
        The output of each $TARGET$ will be written to self.results[$TARGET$]
        '''
        # determine console width
        self.consoleWidth=getConsoleSize()[1]

        if len(targetNames) == 0:
            targetNames=self.options.keys()

        targets={name: self.options[name]
                 for name in targetNames if name in self.options}

        for name, target in sorted(targets.items()):
            options=target.copy()

            if len(extraArgs) > 0:
                options.update(extraArgs)

            result = self._run(name, options)

            # make all result dicts of type accessDict for easy access
            self.results[name] = convertToAccessDicts(result)
            if self.cbResult is not None:
                self.cbResult(name, result)



# -*- coding: utf-8 -*-
'''
  fastmat/core/cmath.pxd
 -------------------------------------------------- part of the fastmat package

  Header file for cython fastmat math library. (types and funcs)

  Author      : wcw
  Introduced  : 2016-10-07
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import cython
import numpy as np
cimport numpy as np

from .types cimport *

cimport numpy as np

################################################## math routines

ctypedef enum OP_MODE:
    MODE_MUL
    MODE_DOTROW

cdef np.float64_t _norm(TYPE_ALL *, intsize)
cdef np.float64_t _normMV(TYPE_ALL[:])
cdef TYPE_ALL _corrMV(TYPE_ALL[:], TYPE_ALL[:])


################################################## complexity estimation
cdef int _findFFTFactors(int, int, int, int)
cpdef intsize _findOptimalFFTSize(intsize, int)
cpdef float _getFFTComplexity(intsize)

################################################## Array generation routines
cpdef np.ndarray _arrZero(int, intsize, intsize, nptype, bint fortranStyle=?)
cpdef np.ndarray _arrEmpty(int, intsize, intsize, nptype, bint fortranStyle=?)

cpdef np.ndarray _arrReshape(np.ndarray, int, intsize, intsize, np.NPY_ORDER)
cpdef bint _arrResize(np.ndarray, int, intsize, intsize, np.NPY_ORDER)

################################################## Array formatting
cpdef np.ndarray _arrCopyExt(np.ndarray, nptype, int)
cpdef np.ndarray _arrForceType(np.ndarray, nptype)
cpdef np.ndarray _arrForceAlignment(np.ndarray, int, bint fortranStyle=?)
cpdef np.ndarray _arrForceTypeAlignment(np.ndarray, nptype, int,
                                        bint fortranStyle=?)

cpdef bint _conjugateInplace(np.ndarray)
cpdef np.ndarray _conjugate(np.ndarray)
cpdef _multiply(np.ndarray, np.ndarray, np.ndarray, ftype, ftype, ftype)
cpdef _dotSingleRow(np.ndarray, np.ndarray, np.ndarray,
                    ftype, ftype, ftype, intsize)



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/DiagBlocks.py
 -------------------------------------------------- part of the fastmat package

  DiagBlocks matrix.


  Author      : sempersn
  Introduced  : 2017-10-06
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------

  TODO:
    - DiagBlocks should simply skip all Zero Matrices (flag them as "None")?
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix


################################################################################
################################################## class DiagBlocks
cdef class DiagBlocks(Matrix):

    ############################################## class methods
    def __init__(self, tenDiags):
        '''Initialize Matrix instance with a list of child matrices'''

        self._numDiagsN = tenDiags.shape[0]
        self._numDiagsM = tenDiags.shape[1]
        self._numDiagsSize = tenDiags.shape[2]

        cdef intsize numN = self._numDiagsN * self._numDiagsSize
        cdef intsize numM = self._numDiagsM * self._numDiagsSize

        self._tenDiags = np.copy(tenDiags)

        dataType = tenDiags.dtype

        # set properties of matrix
        self._initProperties(
            numN, numM, dataType,
            cythonCall=True,
            widenInputDatatype=True
        )

    ############################################## class property override
    cpdef tuple _getComplexity(self):

        return (0, 0)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the forward transform of this matrix'''
        arrRes[:] = np.einsum(
            'nmz,zmk -> znk',
            self._tenDiags,
            arrX.reshape((-1, self._numDiagsM, arrX.shape[1]), order='F')
        ).reshape((-1, arrX.shape[1]), order='F')

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the backward transform of this matrix'''
        arrRes[:] = np.einsum(
            'mnz,zmk -> znk',
            self._tenDiags.conj(),
            arrX.reshape((-1, self._numDiagsM, arrX.shape[1]), order='F')
        ).reshape((-1, arrX.shape[1]), order='F')

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes

        arrRes = np.zeros((self.numN, self.numM), dtype=self.dtype)

        for nn in range(self._numDiagsN):
            for mm in range(self._numDiagsM):
                arrRes[
                    nn *self._numDiagsSize:
                    (nn +1) *self._numDiagsSize,
                    mm *self._numDiagsSize:
                    (mm +1) *self._numDiagsSize
                ] = np.diag(self._tenDiags[nn, mm, :])

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                'size'          : 4,
                TEST.NUM_N      : 32,
                TEST.NUM_M      : 32,
                'mType'        : TEST.Permutation(TEST.ALLTYPES),
                'arr'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType',
                    TEST.SHAPE  : (8, 8, 4)
                }),
                TEST.INITARGS   : (lambda param : [param['arr']()]),
                TEST.OBJECT     : DiagBlocks,
                TEST.NAMINGARGS : dynFormat("(%dx%d) each",
                                            'size', 'size')
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH

        return {
            BENCH.FORWARD: {
                BENCH.FUNC_GEN  : (lambda c: DiagBlocks(
                    np.random.randn(c, c, 64)
                )),
                BENCH.FUNC_SIZE : (lambda c: 64 * c)
            },
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: DiagBlocks(
                    np.random.randn(2 ** c, 2 ** c, c)
                )),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c * c)
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Diagonal Block Matrix (\texttt{fastmat.DiagBlocks})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
For given $n,m \in \N$ this class allows to define a block matrix $\bm M \in
\C^{nm \times nm}$, where each block is a diagonal matrix $\bm D_{ij} \in
\C^{m \times m}$. This obviously allows efficient storage and computations.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the sizes',
                            'n,m = 2,3'
                            '',
                            '# define the diagonals',
                            'd = np.random.randn(',
                            '       n,',
                            '       n,',
                            '       m)',
                            '',
                            '# define the block',
                            '# matrix diagonal-wise',
                            'M = fm.DiagBlocks(d)',
                            caption=r"""
We have randomly drawn the defining elements $\bm d$ from a standard Gaussian
distribution, which results in
\[
    \bm M =
        \left(\begin{array}{cccccc}
                d_{1,1,1} & & & d_{1,2,1} & & \\
                & d_{1,1,2} & & & d_{1,2,2} & \\
                & & d_{1,1,3} & & & d_{1,2,3} \\
                d_{2,1,1} & & & d_{2,2,1} & & \\
                & d_{2,1,2} & & & d_{2,2,2} & \\
                & & d_{2,1,3} & & & d_{2,2,3} \\
              \end{array}\right).\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks',
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD(doc=r"""
$\bm B = \begin{pmatrix}
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} \\
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} \\
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} \\
    \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k} & \bm I_{2^k}
\end{pmatrix}$, $n = 2^{k+2}$ for $k \in \N$""")
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/Parametric.py
 -------------------------------------------------- part of the fastmat package

  Parametric matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.types cimport *
from .core.cmath cimport _dotSingleRow, _conjugateInplace, _arrEmpty


################################################################################
################################################## class Parametric
cdef class Parametric(Matrix):

    ############################################## class properties
    # vecY - Property (read-only)
    # Return the support vector in Y dimension.
    property vecY:
        def __get__(self):
            return self._vecY

    # vecX - Property (read-only)
    # Return the support vector in X dimension.
    property vecX:
        def __get__(self):
            return self._vecX

    # funF - Property (read-only)
    # Return the parameterizing function
    property fun:
        def __get__(self):
            return self._fun

    ############################################## class methods
    def __init__(
        self,
        vecX,
        vecY,
        funF,
        funDtype=None,
        rangeAccess=True
    ):
        '''Initialize Matrix instance'''

        # store flags
        self._rangeAccess = rangeAccess

        # store support vectors, determine dtypes, cast to common dtype
        vecDtype = np.promote_types(vecX.dtype, vecY.dtype)
        self._vecY = vecY.astype(vecDtype)
        self._vecX = vecX.astype(vecDtype)

        # store element function and determine its data type
        # if funDataType is not passed by user, data type is determined
        # upon the function output with the first entries
        self._fun = funF
        self._funDtype = type(funF(self._vecX[0], self._vecY[0])) \
            if funDtype is None else funDtype

        # set properties of matrix
        self._initProperties(
            len(self._vecY),            # numN
            len(self._vecX),            # numM
            self._funDtype,             # data type of matrix
            cythonCall=True,
            forceInputAlignment=True,
            bypassAutoArray=False       # deactivate automatic generation of
                                        # array for transformation bypass. As
                                        # Parametric is always slower than dot
                                        # product with the dense array this
                                        # would otherwise happen always for all
                                        # sizes
        )

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        cdef intsize nn, N = self.numN
        cdef np.ndarray arrRes

        if self._rangeAccess:
            arrRes = self._fun(self._vecX[idx], self._vecY)
        else:
            arrRes = _arrEmpty(1, N, 1, self._info.dtype[0].typeNum)
            for nn in range(N):
                arrRes[nn] = self._fun(self._vecX[idx], self._vecY[nn])

        return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        cdef intsize mm, M = self.numM
        cdef np.ndarray arrRes

        if self._rangeAccess:
            arrRes = self._fun(self._vecX, self._vecY[idx])
        else:
            arrRes = _arrEmpty(1, M, 1, self._info.dtype[0].typeNum)
            for mm in range(M):
                arrRes[mm] = self._fun(self._vecX[mm], self._vecY[idx])

        return arrRes

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._fun(self._vecX[idxM], self._vecY[idxN])

    ############################################## class core methods
    cdef void _core(
        self,
        np.ndarray arrIn,
        np.ndarray arrOut,
        ftype typeIn,
        ftype typeFun,
        ftype typeOut,
        bint backward
    ):

        # determine size of matrices
        cdef intsize mm, nn, numN = self.numN, numM = self.numM

        cdef np.ndarray vecSuppN = self._vecY
        cdef np.ndarray vecSuppM = self._vecX

        # if backward, N and M indexing is swapped and each element must
        # be conjugated
        if backward:
            numN, numM = numM, numN
            vecSuppN, vecSuppM = vecSuppM, vecSuppN

        cdef object val, numSuppN
        cdef np.ndarray vecVal = _arrEmpty(
            1, vecSuppM.shape[0], 1, self._info.dtype[0].typeNum)

        for nn in range(numN):
            # when rangeAccess is allowed, _fun may be called for multiple
            # elements at once
            numSuppN = vecSuppN[nn]
            if self._rangeAccess:
                val = self._fun(
                    numSuppN if backward else vecSuppM,
                    vecSuppM if backward else numSuppN
                )

                if isinstance(val, np.ndarray):
                    # if returned object is an ndarray, assume it to contain
                    # requested data
                    vecVal[:] = val
                else:
                    # otherwise, interpret as scalar value and assign it to all
                    # elements of the requested range
                    for mm in range(numM):
                        vecVal[mm] = val
            elif backward:
                # backward, element-wise
                for mm in range(numM):
                    vecVal[mm] = self._fun(numSuppN, vecSuppM[mm])
            else:
                # forward, element-wirde
                for mm in range(numM):
                    vecVal[mm] = self._fun(vecSuppM[mm], numSuppN)

            # conjugate in backward case
            if backward:
                _conjugateInplace(vecVal)

            # multiply determined vector and accumulate across N. Do a regular
            # scalar product for the first iteration, then accumulate products
            # NOTE: N is actually M in backward case
            _dotSingleRow(arrIn, vecVal, arrOut, typeIn, typeFun, typeOut, nn)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the forward transform of this matrix.'''
        self._core(
            arrX, arrRes, typeX, self._info.dtype[0].fusedType, typeRes, False)

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the backward transform of this matrix.'''
        self._core(
            arrX, arrRes, typeX, self._info.dtype[0].fusedType, typeRes, True)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        arrRes = np.zeros(
            (self.numN, self.numM), dtype=self.dtype)

        for nn in range(self.numN):
            if self._rangeAccess:
                arrRes[nn, :] = self._fun(self._vecX, self._vecY[nn])
            else:
                for mm in range(self.numM):
                    arrRes[nn, mm] = self._fun(self._vecX[mm], self._vecY[nn])

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # define parameters for test
                TEST.NUM_N      : 4,
                TEST.NUM_M      : TEST.Permutation([6, TEST.NUM_N]),
                'typeY'         : TEST.Permutation(TEST.LARGETYPES),
                'typeX'         : TEST.Permutation(TEST.FEWTYPES),
                'rangeAccess'   : TEST.Permutation([False, True]),
                'complexFun'    : TEST.Permutation([False, True]),

                # define arguments for test
                'vecY'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'typeY',
                    TEST.SHAPE  : (TEST.NUM_N, )
                }),
                'vecX'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'typeX',
                    TEST.SHAPE  : (TEST.NUM_M, )
                }),

                # define constructor for test instances and naming of test
                TEST.OBJECT     : Parametric,
                TEST.INITARGS   : (lambda param: [
                    param['vecX'](),
                    param['vecY'](),
                    (TEST.IgnoreFunc(lambda x, y:
                                     x * np.int8(2) - np.complex64(1j) * y)
                     if param['complexFun']
                     else TEST.IgnoreFunc(lambda x, y :
                                          x * np.int8(2) - np.int8(3) * y))
                ]),
                TEST.INITKWARGS : {'rangeAccess': 'rangeAccess'},

                # name the test instances individually to reflect test scenario
                'strC'          : (lambda param:
                                   ('complex' if param['complexFun']
                                    else 'real')),
                'strV'          : (lambda param:
                                   ('vector' if param['rangeAccess']
                                    else 'single')),
                TEST.NAMINGARGS : dynFormat("x:%s,y:%s,%s,%s",
                                            'vecX', 'vecY', 'strV', 'strC'),
                TEST.TOL_POWER  : 4.
            },
            TEST.CLASS: {
                # ignore int8 datatype as there will be overflows
                TEST.IGNORE     : TEST.IgnoreFunc(lambda param: (
                    param['typeX'] == param['typeY'] == np.int8))
            },
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Parametric(
                    np.arange(c).astype(np.double) / c,
                    np.arange(c).astype(np.double) / c,
                    (lambda x, y: np.sin(2 * np.pi * x ** y)))),
                BENCH.FUNC_SIZE : (lambda c: c),
                BENCH.FUNC_STEP : (lambda c: c * 10 ** (1. / 12))
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Parametric(
                    np.arange(2 ** c).astype(np.double) / (2 ** c),
                    np.arange(2 ** c).astype(np.double) / (2 ** c),
                    (lambda x, y: 1.0))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_STEP : (lambda c: c + 1)
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, datatype: Parametric(
                    np.arange(2 ** c).astype(datatype) / (2 ** c),
                    np.arange(2 ** c).astype(datatype) / (2 ** c),
                    (lambda x, y: datatype(1)))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c)
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Parametric Matrix (\texttt{fastmat.Parametric})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
Let $f \D \C^2 \rightarrow \C$ be any function and two vectors $\bm x \in \C^m$
and $\bm y \in \C^n$ such that $(x_j,y_i) \in \Dom(f)$ for $i \in [n]$ and $j
\in [m]$. Then the matrix $\bm F \in \C^{n \times m}$ is defined as

\[F_{i,j} = f(x_j,y_i).\]

This class is not designed to be super fast, but memory efficient. This means,
that everytime the forward or backward projections are called, the elements are
generated according to the specified function on the fly.

\textbf{Hint:} For small dimensions, where the matrix fits into memory, it is
definately more efficient to cast the matrix to a regular Matrix object.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define parameter',
                            '# function for the elements',
                            'def f(x, y):',
                            '    return x ** 2 - y ** 2',
                            '',
                            '# define the input array',
                            '# for the function f',
                            'x = np.linspace(1, 4, 4)',
                            '',
                            '# construct the transform',
                            'F = fm.Parametric(x, x, f)',
                            caption=r"""
This yields
\[f : \C \rightarrow \C\]
\[(x_1,x_2)^T \mapsto x_1^2 - x_2^2\]
\[\bm x = (1,2,3,4)^T\]
\[\bm F = \left(\begin{array}{cccc}
      1 &   3 &  8 & 15 \\
     -3 &   0 &  5 & 12 \\
     -8 &  -5 &  0 &  7 \\
    -15 & -12 & -7 &  0
\end{array}\right)\]"""),
                r"""
We used Cython \cite{para_smith2011cython} to get an efficient implementation in
order to reduce computation time. Moreover, it is generally assumed the the
defined function is able to use row and column broadcasting during evaluation.
If this is not the case, one has to set the flag \texttt{rangeAccess} to
\texttt{False}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix of $\bm \P_{n x n}$ with $n \in \N$
and their support vectors $\bm s_x, \bm s_y \in \R$ each representing a linear
slope along the interval $[0, 1]$.
The defining function varies throughout the benchmarks and is specified
separately for each plot.""",
                DOC.PLOTFORWARD(doc=r"""
$f(x,y) = \exp(2 \pi \cdot x \cdot y)$, $\bm P \in \R^{n \times n}$"""),
                DOC.PLOTFORWARDMEMORY(doc=r"""
$f(x,y) = \exp(2 \pi \cdot x \cdot y)$, $\bm P \in \R^{n \times n}$"""),
                DOC.PLOTSOLVE(doc=r"""
$f(x,y) = \exp(2 \pi \cdot x \cdot y)$, $\bm P \in \R^{n \times n}$"""),
                DOC.PLOTOVERHEAD(doc=r"""
$f(x,y) = \exp(2 \pi \cdot x \cdot y)$, $\bm P \in \R^{n \times n}$"""),
                DOC.PLOTTYPESPEED(doc=r'$f(x,y) = 1$'),
                DOC.PLOTTYPEMEMORY(doc=r'$f(x,y) = 1$')
            ),
            DOC.BIBLIO(
                para_smith2011cython=DOC.BIBITEM(
                    r"""
Stefan Behnel, Robert Bradshaw, Craig Citro, Lisandro Dalcin,
Dag Sverre Seljebotn and Kurt Smith,""",
                    r'Cython: The Best of Both Worlds',
                    r'Computing in Science and Engineering, Volume 13, 2011.')
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/inspect/documentation.py
 -------------------------------------------------- part of the fastmat package

  Implementations for class testing (quality management).


  Author      : wcw
  Introduced  : 2017-07-19
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import inspect
import os
import sys
from datetime import datetime

from .common import *
from .benchmark import BENCH, Benchmark

modInspect = [sys.modules['fastmat.inspect.%s' %(name, )]
              for name in ['common', 'benchmark']]


################################################## DOC class
class DOC(NAME):
    CAPTION     = 'caption'
    TITLE       = 'title'
    XLABEL      = 'xLabel'
    YLABEL      = 'yLabel'
    DATASETS    = 'datasets'
    LEGEND      = 'legend'

    MAKRO       = 'makro'
    MAKRO_SPEED     = r'\speed'
    MAKRO_MEMORY    = r'\mem'
    MAKRO_OVERHEAD  = r'\overhead'
    MAKRO_TYPES     = r'\dtypes'

    OUTPATH     = 'outPath'
    FILENAME    = 'filename'
    TARGET      = 'target'
    COLUMN      = 'col'

    class __DOC__(object):
        def __init__(self, *args):
            self.items = args

        def __str__(self):
            return self.get()

        def __call__(self, **options):
            return self.get(**options)

        def get(self, **options):
            return os.linesep.join([''] + [
                ii.get(**options) if isinstance(ii, DOC.__DOC__)
                else str(ii).strip()
                for ii in self.items] + [''])

    class SECTION(__DOC__):
        def __init__(self, title, *contents):
            self.items = [r"\section{%s}" %(title)]
            self.items.extend(contents)

    class SUBSECTION(__DOC__):
        def __init__(self, title, *contents):
            self.items = [r"\subsection{%s}" %(title)]
            self.items.extend(contents)

    class SUBSUBSECTION(__DOC__):
        def __init__(self, title, *contents):
            self.items = [r"\subsubsection{%s}" %(title)]
            self.items.extend(contents)

    class BIBITEM(__DOC__):
        def __init__(self, author, title, source):
            self.items = [author, r"\emph{%s}" %(title), source]

    class BIBLIO(__DOC__):
        def __init__(self, **entries):
            self.items = [r"\begin{thebibliography}{9}"]
            for tag, entry in entries.items():
                self.items.extend([r"\bibitem{%s}" %(tag), entry])

            self.items.append(r"\end{thebibliography}")

    class SNIPPET(__DOC__):
        def __init__(self, *code, **opts):
            # output listing environment (with optional title)
            self.items = [
                r"\begin{snippet}%s" %("[%s]" %(opts[DOC.TITLE])
                                       if DOC.TITLE in opts else ""),
                r"\begin{lstlisting}[language=Python]"]

            # now add the contents
            self.items.extend(code if isinstance(code, (list, tuple))
                              else [code])

            # output caption if needed
            self.items.append(r"\end{lstlisting}")
            if DOC.CAPTION in opts:
                self.items.append(opts[DOC.CAPTION])

            # ... and finish the environment
            self.items.append(r"\end{snippet}")

    class PLOT(__DOC__):
        def __init__(self, benchName, **options):
            self.benchName = benchName
            self.params = options

        def get(self, **options):
            # merge options and parameters
            params = mergeDicts(self.params, options)
            params.setdefault(DOC.CAPTION, '')
            params.setdefault(DOC.TITLE, '')

            # check if data file actually exists
            benchmark = params.get(NAME.BENCHMARK, None)
            if benchmark is None or self.benchName not in benchmark.options:
                return ""

            outPath = params.get(DOC.OUTPATH, '')
            addVersionTag = params.get('addVersionTag', True)
            filename = benchmark.getFilename(self.benchName, path=outPath,
                                             addVersionTag=addVersionTag)
            params[DOC.FILENAME] = filename
            params[DOC.TARGET] = self.benchName

            # if result file exists, determine its age and compare it to the
            # age of the modules involved in the benchmark target.
            # This also involves all classes the benchmark target depends on
            if os.path.isfile(filename):
                # determine age of result file
                fileAge = os.path.getmtime(filename)

                # now determine a list of all modules containing code used by
                # the particular class and the quality management sections
                modules = [module
                           for module in inspect.getmro(
                               benchmark.target.__class__)
                           if module.__module__ != '__builtin__']
                modules.extend(modInspect)

                # now determine the newest all modules found as moduleage
                def getModuleAge(module):
                    try:
                        return os.path.getmtime(inspect.getfile(module))
                    except TypeError:
                        return 0

                moduleAge = max([getModuleAge(module) for module in modules])
            else:
                fileAge, moduleAge = -1, 0

            # the file is considered up-to-date if fileAge does exceed moduleAge
            # if necessary, rerun benchmark and update file
            if fileAge < moduleAge:
                if moduleAge == 0:
                    print("Output '%s' not found for target '%s'. Rebuild." %(
                        filename, self.benchName))
                else:
                    def strTime(timestamp):
                        return datetime.fromtimestamp(timestamp).strftime(
                            "%Y-%m-%d %H:%M:%S")

                    print(("Output '%s' outdated for target '%s'. " +
                           "(%s < %s). Rebuild.") %(filename, self.benchName,
                                                    strTime(fileAge),
                                                    strTime(moduleAge)))

                benchmark.run(self.benchName)
                benchmark.saveResult(self.benchName, outPath=outPath,
                                     addVersionTag=addVersionTag)

            # output header
            self.items = [r'\tikzpicturedependsonfile{%(filename)s}' %(params),
                          r'%(makro)s{' %(params)]

            # now do the actual plotting
            for dataset in params[DOC.DATASETS]:
                self.items.append(r"""
\addplot table[x=numN, col sep=comma,y=%s]{%s};""" %(dataset,
                                                     params[DOC.FILENAME]))

            # now output the legend
            self.items.append(r'\legend{%s};' %(",".join(params[DOC.LEGEND])))

            # output footer
            self.items.extend([r'}{%s}{%s}' %(params.get(DOC.TITLE),
                                              params.get(DOC.YLABEL))])

            return super(DOC.PLOT, self).get()

    ############################################## meta targets for plotting
    class PLOTSPEED(PLOT):
        def __init__(self, name, **options):
            super(DOC.PLOTSPEED, self).__init__(name, **mergeDicts(
                {DOC.DATASETS   : ['fastmatMin', 'numpyMin'],
                 DOC.LEGEND     : ['fastmat', 'numpy'],
                 DOC.MAKRO      : DOC.MAKRO_SPEED,
                 DOC.YLABEL     : 'Runtime [s]'}, options))

    class PLOTMEMORY(PLOT):
        def __init__(self, name, **options):
            super(DOC.PLOTMEMORY, self).__init__(name, **mergeDicts(
                {DOC.TITLE      : 'Memory consumption',
                 DOC.DATASETS   : ['fastmatMem', 'numpyMem'],
                 DOC.LEGEND     : ['fastmat', 'numpy'],
                 DOC.MAKRO      : DOC.MAKRO_MEMORY,
                 DOC.YLABEL     : 'Memory [kB]'}, options))

    class PLOTTYPES(PLOT):
        def __init__(self, name, **options):
            typeList = options.get(
                'typelist', ['c32', 'c64', 'f32', 'f64', 'i08', 'i32', 'i64'])
            col = options.get(DOC.COLUMN)
            super(DOC.PLOTTYPES, self).__init__(name, **mergeDicts(
                {DOC.DATASETS   : ["%s%s" %(str, col) for str in typeList],
                 DOC.LEGEND     : typeList,
                 DOC.MAKRO      : DOC.MAKRO_TYPES}, options))

    ############################################## plot targets using metas
    class PLOTFORWARD(PLOTSPEED):
        def __init__(self, **options):
            super(DOC.PLOTFORWARD, self).__init__(BENCH.FORWARD, **mergeDicts(
                {DOC.TITLE      : 'Forward Projection'}, options))

    class PLOTSOLVE(PLOTSPEED):
        def __init__(self, **options):
            super(DOC.PLOTSOLVE, self).__init__(BENCH.SOLVE, **mergeDicts(
                {DOC.TITLE      : 'Solving a LSE'}, options))

    class PLOTFORWARDMEMORY(PLOTMEMORY):
        def __init__(self, **options):
            super(DOC.PLOTFORWARDMEMORY, self).__init__(
                BENCH.FORWARD, **mergeDicts({DOC.TITLE: 'Class Memory usage'},
                                            options))

    class PLOTTYPESPEED(PLOTTYPES):
        def __init__(self, **options):
            super(DOC.PLOTTYPESPEED, self).__init__(BENCH.DTYPES, **mergeDicts(
                {DOC.TITLE      : 'Datatype Impact (Speed)',
                 DOC.COLUMN     : 'Min',
                 DOC.YLABEL     : 'Runtime [s]'}, options))

    class PLOTTYPEMEMORY(PLOTTYPES):
        def __init__(self, **options):
            super(DOC.PLOTTYPEMEMORY, self).__init__(
                BENCH.DTYPES,
                **mergeDicts({DOC.TITLE      : 'Datatype Impact (Memory)',
                              DOC.COLUMN     : 'Mem',
                              DOC.YLABEL     : 'Memory [kB]'}, options))

    ############################################## stand-alone plot targets
    class PLOTOVERHEAD(PLOTSPEED):
        def __init__(self, **options):
            super(DOC.PLOTOVERHEAD, self).__init__(BENCH.OVERHEAD, **mergeDicts(
                {DOC.TITLE      : 'Transform Runtime',
                 DOC.DATASETS   : ['forwardMin', 'backwardMin'],
                 DOC.LEGEND     : ['forward', 'backward']}, options))

    class PLOTPERFORMANCE(PLOTSPEED):
        def __init__(self, **options):
            super(DOC.PLOTPERFORMANCE, self).__init__(
                BENCH.PERFORMANCE, **mergeDicts({
                    DOC.TITLE      : 'Runtime Performance',
                    DOC.DATASETS   : ['perfMin'],
                    DOC.LEGEND     : ['runtime']}, options))


################################################################################
################################################## Documentation function
def Documentation(targetClass, **options):

    # the test target is a fastmat class to be instantiated in the runners
    if not inspect.isclass(targetClass):
        raise ValueError("target in init of Runner must be a class type")

    # fetch benchmark options from options
    benchmarkOptions = options.pop('benchmarkOptions', {})

    # generate a benchmark instance for access to classes benchmarks
    # also use this instance to access documentation options
    # if no options can be retrieved, exit.
    if DOC.BENCHMARK not in options:
        options[DOC.BENCHMARK] = Benchmark(targetClass,
                                           extraOptions=benchmarkOptions)

    benchmark = options[DOC.BENCHMARK]
    doc = getattr(benchmark.target, '_getDocumentation', lambda: "")()
    return (doc.get(**options) if isinstance(doc, DOC.__DOC__)
            else str(doc))



# -*- coding: utf-8 -*-
'''
  fastmat/Matrix.pxd
 -------------------------------------------------- part of the fastmat package

  Header file for Matrix base class (structural description).


  Author      : wcw
  Introduced  : 2016-09-24
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import cython
import numpy as np
cimport numpy as np

from .core.types cimport *

################################################## class MatrixProfile
cdef class FastmatFlags:
    cpdef public bint bypassAllow
    cpdef public bint bypassAutoArray

################################################## class MatrixProfile
cdef class MatrixCalibration:
    cdef public float offsetForward
    cdef public float offsetBackward
    cdef public float gainForward
    cdef public float gainBackward


################################################## class MatrixProfile
ctypedef struct PROFILE_s:
    float complexity
    float overhead
    float overheadNested
    float effort
    float effortNested

cdef bint profileUpdate(PROFILE_s *, intsize M, bint, PROFILE_s *, PROFILE_s *)

################################################## class Matrix
cdef class Matrix:

    ############################################## class variables
    cdef public np.ndarray  _array               # ndarray dense representation
    cdef public np.ndarray  _arrayH              # backward dense representation
    cdef public tuple   _content                 # nested fastmat matrices

    cdef public Matrix  _gram                    # cache for gram matrix
    cdef public Matrix  _normalized              # cache for normalized matrix
    cdef public object  _largestEV               # cache for largestEV
    cdef public object  _largestSV               # cache for largestSV
    cdef public Matrix  _T                       # cache for transpose matrix
    cdef public Matrix  _H                       # cache for adjunct matrix
    cdef public Matrix  _conj                    # cache for conjugate matrix

    cdef intsize     _numN                       # row-count of matrix
    cdef intsize     _numM                       # column-count of matrix
    cdef INFO_ARR_s  _info                       # fields representing matrix
    #                                            #  dtype - data type descriptor
    #                                            #  numN  - row count
    #                                            #  numM  - column count
    #                                            #  nDim  - dimension count
    cdef bint   _cythonCall                      # use _C() transforms in class
    cdef bint   _forceInputAlignment             # force alignment of input data
    #                                            # to be F-contiguous
    cdef bint   _useFortranStyle                 # if true, select Fortran style
    cdef bint   _widenInputDatatype              # widen input data type upfront
    cdef str    _tag                             # Description of matrix
    cdef bint   _bypassAllow                     # if true, transform may be
    #                                            # bypassed based on runtime
    #                                            # estimation decision
    cdef bint   _bypassAutoArray                 # if true, a dense array repre-
    #                                            # sentation for bypassing a
    #                                            # transform will automatically
    #                                            # be generated when required

    cdef public object  _forwardReferenceMatrix  # ndarray representing Matrix
    #                                            # reference

    ############################################## class profiling
    cdef PROFILE_s _profileBypassFwd
    cdef PROFILE_s _profileBypassBwd
    cdef PROFILE_s _profileForward
    cdef PROFILE_s _profileBackward

    ############################################## class implementation methods
    cpdef np.ndarray _getArray(self)
    cpdef np.ndarray _getCol(self, intsize)
    cpdef np.ndarray _getRow(self, intsize)
    cpdef object _getItem(self, intsize, intsize)
    cpdef object _getLargestEV(self, intsize, float, float, bint)
    cpdef object _getLargestSV(self, intsize, float, float, bint)
    cpdef Matrix _getGram(self)
    cpdef Matrix _getNormalized(self)
    cpdef Matrix _getT(self)
    cpdef Matrix _getH(self)
    cpdef Matrix _getConj(self)

    ############################################## computation profiling
    cpdef tuple _getComplexity(self)
    cdef void _initProfiles(self)
    cpdef _exploreNestedProfiles(self)

    cpdef _forwardC(self, np.ndarray, np.ndarray, ftype, ftype)
    cpdef _backwardC(self, np.ndarray, np.ndarray, ftype, ftype)
    cpdef np.ndarray _forward(self, np.ndarray)
    cpdef np.ndarray _backward(self, np.ndarray)

    ############################################## class interface methods
    cpdef np.ndarray forward(self, np.ndarray)
    cpdef np.ndarray backward(self, np.ndarray)
    cpdef np.ndarray getArray(self)

    ############################################## class reference
    cpdef np.ndarray _reference(self)
    cpdef np.ndarray reference(self)

################################################## Hermitian,Conjugate,Transpose
cdef class Hermitian(Matrix):
    cdef public Matrix _nested                   # nested fastmat baseclass

cdef Matrix getConjugate(Matrix)
cdef class Conjugate(Matrix):
    cdef public Matrix _nested                   # nested fastmat baseclass

cdef class Transpose(Hermitian):
    cdef public Matrix _nestedConj               # nested fastmat baseclass
